/++
Модуль, определяющий класс $(LINK2 #DDDМенеджер, DDDМенеджер), отвечающий за доступ к
ресурсам, обработку событий, параметры сцены и её визуализацию.

Экземпляр класса присутствует и инициализируется в модуле $(LINK2 package.html, ddd), 
его не требуется создавать самостоятельно.
+/
module ddd.ddd_manager;

import ddd.vis, ddd.mesh, ddd.resources, ddd.camera, ddd.mesto, ddd.vector, ddd.quaternion, ddd.texture;
import ddd.skelet, ddd.infopanel;
import ddd.zavisimost.vis_gl, ddd.zavisimost.sobytija_SDL2, ddd.zavisimost.shrift_SDL2;
import my_utils.journal;

private enum имя_модуля = "ddd_менеджер";

/++
Класс, отвечающий за доступ к
ресурсам, обработку событий, параметры сцены и её визуализацию.

Часть данных-членов, относящихся к визуализации сцены, объявлена публично для 
более удобного доступа к построению и изменению этой сцены.
+/
class DDDМенеджер {

    ///Объект $(LINK2 vis.html#Визуализатор, Визуализатор), отвечающий за визуализацию сцены.
    public Визуализатор визуализатор;

    ///Объект $(LINK2 camera.html#Камера, Камера), являющийся камерой сцены по-умолчанию.
    public Камера камера;

    ///Объект $(LINK2 mesto.html#Место, Место), являющийся корнем дерева построения сцены.
    public Место начальное_место;
    ///Объект $(LINK2 mesto.html#Место, Место), в который можно вкладывать объекты, изображающие фон или нечто далёкое.
    public Место место_фона;

    /++
    Объект $(LINK2 mesto.html#Место, Место), к которому присоединена камера по-умолчанию. 
    Через него рекомендуется задавать расположение камеры.
    +/
    public Место место_камеры;

    private ИменаРесурсов имена_ресурсов;
    private ЗагруженныеМеши загруженные_меши;
    private ЗагруженныеСкелеты загруженные_скелеты;
    private ЗагруженныеТекстуры загруженные_текстуры;
    private ЗагруженныеTTFШрифты загруженные_ttf_шрифты;
    private ЗагруженныеИнфопанели загруженные_инфопанели;


    private ОбработчикСобытий обработчик_событий = null;
    private void function() обновление_сцены = null;
    private Инфопанель[string] инфопанели;
    
    /++
    Конструктор, в котором создаются объекты, содержащие загруженные ресурсы,
    объект визуализатора, камера и три объекта типа $(LINK2 mesto.html#Место, Место).

    Params:
        каталоги_ресурсов = Массив строк, содержащий пути к каталогам с ресурсами.
    +/
    this(string[] каталоги_ресурсов) {
        имена_ресурсов = new ИменаРесурсов(каталоги_ресурсов);
        загруженные_меши = new ЗагруженныеМеши(имена_ресурсов);
        загруженные_скелеты = new ЗагруженныеСкелеты(имена_ресурсов);
        загруженные_текстуры = new ЗагруженныеТекстуры(имена_ресурсов);
        загруженные_ttf_шрифты = new ЗагруженныеTTFШрифты(имена_ресурсов);
        загруженные_инфопанели = new ЗагруженныеИнфопанели(имена_ресурсов, this);

        в_журнал(имя_модуля, УровниЖурнала.Отладка, "Сейчас будем создавать визуализатор");
        визуализатор = new Визуализатор_gl(this);
        начальное_место = new Место("Начальное место");
        место_фона = new Место("Место фона");
        камера = new Камера("Камера по-умолчанию");
        //место_камеры = начальное_место.создать_ребёнка("Место камеры", 
        место_камеры = new Место("Место камеры"); 
        место_камеры.задать_позицию(Вектор3(0,0,3));
        место_камеры.задать_ориентацию(Кватернион.НОЛЬ);
        в_журнал(имя_модуля, УровниЖурнала.Отладка, место_камеры.toString());
        место_камеры.присоединить_объект(камера);
    }
    
    /++
    Позволяет узнать, был ли задан обработчик событий.

    Returns:
        true, если обработчик событий не был задан, false в противном случае.
    +/
    bool обработка_событий_отсутствует() {
        return обработчик_событий is null;
    }    

    /++
    Выполняет обработку событий для текущего кадра.

    В первом кадре, если пользователь не назначал менеджеру собственный обработчик событий,
    создаётся экземпляр класса $(LINK2 sobytija.html#ОбработчикСобытий, ОбработчикСобытий), который 
    и используется в дальнейшем.

    Returns:
        Результат метода $(B обработка()) обработчика событий. Если это true - то программа 
        будет завершена.
    +/
    bool обработка_событий() {
        if (обработчик_событий is null)
            обработчик_событий = new ОбработчикСобытий();
        return обработчик_событий.обработка();
    }

    /+
    Метод позволяет пользователю определить и использовать собственный обработчик событий,
    для реализации обработки событий (мыши или клавиатуры), отличной от заданной по-умолчанию.

    Params:
        обработчик = собственный обработчик событий пользователя. Должен быть подклассом
        класса $(LINK2 sobytija.html#ОбработчикСобытий, ОбработчикСобытий).
    +/
    void задать_обработчик_событий(ОбработчикСобытий обработчик) {
        обработчик_событий = обработчик;
    }

    /++
    Метод позволяет пользователю задать функцию обновления сцены.

    Params:
        функция_обновления_сцены = Функция обновления сцены, которую визуализатор будет вызывать 
        в каждом кадре.
    +/
    void задать_функцию_обновления_сцены(void function() функция_обновления_сцены) {
        обновление_сцены = функция_обновления_сцены;
    }

    /++
    Метод, вызываемый визуализатором в каждом кадре. Если не была задана функция обновления сцены,
    метод ничего не делает.
    +/
    void обновить_сцену() {
        if (обновление_сцены !is null)
            обновление_сцены();
    }

    /++
    Метод, возвращающий меш из загруженных ресурсов.

    Params:
        имя_меша = Имя файла меша (без расширения).
    +/
    Меш получить_меш(string имя_меша) {
        return (загруженные_меши.получить(имя_меша));
    }


    /++
    Метод, возвращающий скелет из загруженных ресурсов.


    Params:
        имя_скелета = Имя файла скелета (без расширения).
    +/
    Скелет получить_скелет(string имя_скелета) {
        return (загруженные_скелеты.получить(имя_скелета));
    }
    

    /++
    Метод, возвращающий текстуру из загруженных ресурсов.


    Params:
        имя_текстуры = Имя графического файла, используемого для текстуры (без расширения).
    +/
    Текстура получить_текстуру(string имя_текстуры) {
        return (загруженные_текстуры.получить(имя_текстуры));
    }
    

    /++
    Метод, возвращающий $(LINK2 shrift.html#TTFШрифт, TTFШрифт) из загруженных ресурсов.


    Params:
        имя_шрифта = Имя файла со шрифтом (без расширения).
    +/
    TTFШрифт получить_ttf_шрифт(string имя_шрифта) {
        return (загруженные_ttf_шрифты.получить(имя_шрифта));
    }
    

    /++
    Метод, возвращающий инфопанель из загруженных ресурсов.


    Params:
        имя_инфопанели = Имя файла с описанием инфопанели (без расширения).
    +/
    Инфопанель получить_инфопанель(string имя_инфопанели) {
        return (загруженные_инфопанели.получить(имя_инфопанели));
    }
    

    /++
    Метод добавляющий инфопанель на экран.


    Params:
        инфопанель = Объект класса $(LINK2 infopanel.html#Инфопанель, Инфопанель), 
        который требуется вывести на экран.
    +/
    void добавить_инфопанель(Инфопанель инфопанель) {
        инфопанели[инфопанель.получить_имя] = инфопанель;
    }


    /++
    Метод, вызываемый визуализатором в каждом кадре, по очереди отрисовывающий все инфопанели. 
    +/
    void отрисовка_инфопанелей() {
        foreach (имя, инфопанель; инфопанели) {
            инфопанель.рисовать();
        }
    }
}
