/++
Модуль, определяющий класс $(LINK2 #Инфопанель, Инфопанель), который позволяет 
добавлять на экран двумерные перекрывающие сцену панели, на них можно
размещать какую-либо информацию, например, картинки, текст или анимацию.

Также определён класс ресурсов $(LINK2 #ЗагруженныеИнфопанели, ЗагруженныеИнфопанели), позволяющий получать объекты 
$(LINK2 #Инфопанель, Инфопанель) из файлов-ресурсов json-формата с расширением ".ip". 
Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты класса $(LINK2 #Инфопанель, Инфопанель), созданные из json-файлов, через него.
+/
module ddd.infopanel;

import std.conv, std.string, std.json, std.file;
import my_utils.journal;
import ddd.resources, ddd.ddd_manager, ddd.types, ddd.okno, ddd.element_infopaneli;
import ddd.zavisimost.graphika_gl;

enum имя_модуля = "инфопанель";
enum ТИП_РЕСУРСА = "инфопанель";

/++
Типы координат местоположения инфопанели:
$(UL  
$(LI  $(B Абсолютно) - координаты в пикселях) 
$(LI  $(B Относительно) - x относительно ширины окна, y относительно высоты окна)  
$(LI  $(B ОтносительноX) - x,y относительно ширины окна)  
$(LI  $(B ОтносительноY) - x,y относительно высоты окна)  
)
+/
enum ТипКоординат {Абсолютно, Относительно, ОтносительноX, ОтносительноY};

private enum ТипКоординат[string] ПреобразовательТипаКоординат= [
    "Абсолютно":     ТипКоординат.Абсолютно, 
    "Относительно":  ТипКоординат.Относительно, 
    "ОтносительноX": ТипКоординат.ОтносительноX, 
    "ОтносительноY": ТипКоординат.ОтносительноY
    ];

/++
Инфопанель в json-файле:
$(UL  
$(LI  $(B имя) (строковый) - имя инфопанели в программе) 
$(LI  $(B тип_координат) (строковый) - как именно задаются координаты расположения инфопанели, 
    см. $(LINK2 #ТипКоординат, ТипКоординат))  
$(LI  $(B координаты) (массив из 4-х значений типа float) - координаты инфопанели: [
координата x левой границы, координата y нижней границы, ширина, высота
])  
$(LI  $(B элементы) - массив объектов $(LINK2 element_infopaneli.html#ЭлементИнфопанели, ЭлементИнфопанели))  
)
+/
enum JSON_TYPE[string] Инфопанель_JSON = [
    "имя": JSON_TYPE.STRING,
    "тип_координат": JSON_TYPE.STRING,
    "координаты": JSON_TYPE.ARRAY,
    "элементы": JSON_TYPE.ARRAY
    ];

//enum ИнфопанельКоординатаJSON = JSON_TYPE.FLOAT;
//enum Overlay_Elements_JSON = JSON_TYPE.FLOAT;

/++
Класс, позволяющий
добавлять на экран двумерные перекрывающие сцену панели (изначально прозрачные), 
на них можно размещать какую-либо информацию, например, картинки, текст или анимацию.

В инфопанель добавляются объекты класса $(LINK2 element_infopaneli.html#ЭлементИнфопанели, ЭлементИнфопанели) 
или его подклассов. В контейнере инфопанели они содержатся в виде ассоциативного массива, т.е.
имена добавленных элементов должны быть уникальны в пределах всей инфопанели.

Во все конструкторы необходимо передавать ссылку на менеджер программы, т.к. он требуется для 
создания и визуализации текста.
+/
class Инфопанель {    
    private {
        DDDМенеджер менеджер;
        bool включено = true;
        ТипКоординат тип_координат = ТипКоординат.Абсолютно;
        float относительный_x, относительный_y, относительная_ширина, относительная_высота;
        int x,y, ширина, высота;
        string имя;
        ЭлементИнфопанели контейнер;
    }

    /++
    Конструктор, принимающий имя и ссылку на DDDМенеджер этой программы.

    Params:
        имя0 = имя новой инфопанели
        менеджер0 = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)
    +/
    this(string имя0, DDDМенеджер менеджер0) {
        /+
    При использовании этого конструктора (т.е. без использования координат) будет создана
    инфопанель размером во весь экран.
        +/
        имя = имя0;
        менеджер = менеджер0;
        инициализация_контейнера();
    }

    /++
    Конструктор, принимающий имя, ссылку на DDDМенеджер этой программы и координаты расположения.

    Отрицательные координаты x0 и/или y0 задают положение от конца экрана (x0 от правого края, y0 от верхнего).


    Params:
        имя0 = имя новой инфопанели
        менеджер0 = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)
        тип0 = тип передаваемых координат, см. $(LINK2 #ТипКоординат, ТипКоординат)
        x0 = координата x левой границы инфопанели
        y0 = координата y нижней границы инфопанели
        ширина0 = ширина инфопанели
        высота0 = высота инфопанели
    +/
    this(string имя0, DDDМенеджер менеджер0, ТипКоординат тип0, float x0, float y0, float ширина0, float высота0) {
        имя = имя0;
        менеджер = менеджер0;
        инициализация_контейнера();
        задать_координаты(тип0, x0, y0, ширина0, высота0);
    }

    private void инициализация_контейнера() {
        контейнер = new ЭлементИнфопанели(имя ~ "_контейнер", true, 0,0,1,1);
    }

    /++
    Функция, задающая координаты расположения этой инфопанели.


    Отрицательные координаты x0 и/или y0 задают положение от конца экрана (x0 от правого края, y0 от верхнего).


    Params:
        тип0 = тип передаваемых координат, см. $(LINK2 #ТипКоординат, ТипКоординат)
        x0 = координата x левой границы инфопанели
        y0 = координата y нижней границы инфопанели
        ширина0 = ширина инфопанели
        высота0 = высота инфопанели
    +/
    public void задать_координаты(ТипКоординат тип0, float x0, float y0, float ширина0, float высота0) {
        тип_координат = тип0;
        auto рзм_окна = размеры_окна();
        int ширина_окна = рзм_окна.x;
        int высота_окна = рзм_окна.y;
        switch(тип0) {
            case ТипКоординат.Абсолютно:
                this.x= to!(int)(x0>=0? x0 : (ширина_окна + x0));
                this.y= to!(int)(y0>=0? y0 : (высота_окна + y0));
                this.ширина = to!(int)(ширина0);
                this.высота = to!(int)(высота0);
                относительный_x = this.x / ширина_окна;
                относительная_ширина = ширина0 / ширина_окна;
                относительный_y = this.y / высота_окна;
                относительная_высота = высота0 / высота_окна;
                break;
            case ТипКоординат.Относительно:
                относительный_x = x0>=0? x0 : 1+x0;
                относительный_y = y0>=0? y0 : 1+y0;
                this.x = to!(int)(относительный_x * ширина_окна);
                this.y = to!(int)(относительный_y * высота_окна);
                this.ширина = to!(int)(ширина_окна * ширина0);
                this.высота = to!(int)(высота_окна * высота0);
                относительная_ширина = ширина0;
                относительная_высота = высота0;
                break;
            case ТипКоординат.ОтносительноX:
                относительный_x = x0>=0? x0 : 1+x0;
                this.x = to!(int)(относительный_x * ширина_окна);
                относительная_ширина = ширина0;
                this.ширина = to!(int)(ширина_окна * ширина0);
                this.y = to!(int)(y0 * this.x);
                this.высота = to!(int)(высота0 * this.ширина);
                относительный_y = this.y / высота_окна;
                относительная_высота = this.высота / высота_окна;
                break;
            case ТипКоординат.ОтносительноY:
                относительный_y = y0>=0? y0 : 1+y0;
                this.y = to!(int)(высота_окна * относительный_y);
                относительная_высота = высота0;
                this.высота = to!(int)(высота_окна * высота0);
                this.x = to!(int)(x0 * this.y);
                this.ширина = to!(int)(ширина0 * this.высота);
                относительный_x = this.x / ширина_окна;
                относительная_ширина = this.ширина / ширина_окна;
                break;
            default:
                return;
        }
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, this.toString());
        контейнер.рассчитать_абсолютные_координаты(this.ширина, this.высота);
    } 

    /++
    Вывод этой инфопанели в строковое представление.
    +/
    public override string toString() {
        return format("%s включено=%s, x=%s, y=%s, ширина=%s, высота=%s отн_x=%s, отн_y=%s, отн_ш=%s, отн_в=%s",
                      имя, this.включено, this.x, this.y, this.ширина, this.высота, 
                      относительный_x, относительный_y, относительная_ширина, относительная_высота); 
    }

    /// Функция, позволяющая включить или выключить видимость этой инфопанели.
    public void задать_видимость(bool включено0) {
        включено = включено0;
    }

    /// Функция, возвращающая значение видимости этой инфопанели.
    public @property bool получить_видимость() {
        return включено;
    }

    /// Функция, позволяющая переключить видимость этой инфопанели.
    public void поменять_видимость() {
        включено = !включено;
    }

    /// Функция, возвращающая значение имени этой инфопанели.
    public @property string получить_имя() {
        return имя;
    }

    /++
    Присоединение элемента к этой инфопанели.

    Params:
        элемент = присоединяемый элемент
    +/
    public void присоединить_элемент(ЭлементИнфопанели элемент) {
        контейнер.присоединить_элемент(элемент);
    }

    /++
    Отсоединение элемента от этой инфопанели.


    Params:
        элемент = отсоединяемый элемент

    Returns:
        Отсоединённый элемент, если он присуствовал в этой инфопанели, null, если нет.
    +/
    public ЭлементИнфопанели отсоединить_элемент(ЭлементИнфопанели элемент) {
        return контейнер.отсоединить_элемент(элемент);
    }

    /++
    Отсоединение элемента от этой инфопанели по его имени.


    Params:
        имя_элемента = имя отсоединяемого элемента


    Returns:
        Отсоединённый элемент, если он присуствовал в этой инфопанели, null, если нет.
    +/
    public ЭлементИнфопанели отсоединить_элемент(string имя_элемента) {
        return контейнер.отсоединить_элемент(имя_элемента);
    }

    /++
    Получение ссылки на элемент этой инфопанели по его имени.


    Params:
        имя_элемента = имя искомого элемента


    Returns:
        Искомый элемент, если он присуствовал в этой инфопанели, null, если нет.
    +/
    public ЭлементИнфопанели получить_элемент(string имя_элемента) {
        return контейнер.получить_элемент(имя_элемента);
    }

    static public ТипКоординат преобразовать_тип_координат(string тип) {
        // По-умолчанию ТипКоординат.Абсолютно
        return ПреобразовательТипаКоординат.get(тип, ТипКоординат.Абсолютно);
    }

    /++
    Вывод этой инфопанели на экран. Если видимость отключена, то функция не выполняет никаких действий.
    +/
    public void рисовать() {
        if (!включено) return;
        gl_подготовить_рисование_инфопанели(x, y, ширина, высота);
        контейнер.рисовать();
    }

    /++
    Функция генерации новой инфопанели (со всеми присоединёнными элементами) из JSON-объекта.

    Params:
        json_значение0 = JSON-объект с информацией об инфопанели (см. $(LINK2 #Инфопанель_JSON, Инфопанель_JSON)).
        менеджер0 = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)

    Returns:
        Созданная инфопанель, если генерация прошла успешно, null в противном случае.
    +/
    static public Инфопанель из_JSON(JSONValue json_значение0, DDDМенеджер менеджер0) {
        JSONValue[string] json_значение1;
        string имя;
        JSONValue[] координаты;
        ТипКоординат тип_координат = ТипКоординат.Абсолютно;
        Инфопанель результат;
        try {
            json_значение1 = json_значение0.object;
            имя = json_значение1["имя"].str;
            координаты = json_значение1["координаты"].array;
        }
        catch (Throwable)
            return null;
        try
            тип_координат = ПреобразовательТипаКоординат[json_значение1["тип_координат"].str];
        catch (Throwable)
            {}
        try
            результат = new Инфопанель(имя, менеджер0, тип_координат, координаты[0].floating, 
                            координаты[1].floating, координаты[2].floating, координаты[3].floating);
        catch (Throwable)
            return null;
        if ("элементы" in json_значение1)
            if (json_значение1["элементы"].type == JSON_TYPE.ARRAY) {
                foreach (JSONValue json_значение2; json_значение1["элементы"].array) {
                    ЭлементИнфопанели элемент = ЭлементИнфопанели.из_JSON(json_значение2, менеджер0);
                    if (элемент) {
                        результат.присоединить_элемент(элемент);
                    }
                }
            }
        return результат;
    }    
    
}
    
/++
Класс-контейнер ресурсов объектов класса $(LINK2 #Инфопанель, Инфопанель).


Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты класса $(LINK2 #Инфопанель, Инфопанель), созданные из json-файлов, через него.
+/
class ЗагруженныеИнфопанели : ЗагруженныеРесурсы!(Инфопанель) {
    
    private DDDМенеджер менеджер;

    /++
    Конструктор, принимающий ссылку на объект типа $(LINK2 #ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов, и на менеджер программы.

    Params:
        имена_ресурсов0 = объект типа $(LINK2 resources.html#ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов
        менеджер0 = менеджер программы, объект класса $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)
    +/
    this(ИменаРесурсов имена_ресурсов0, DDDМенеджер менеджер0) {
        super(имена_ресурсов0);
        this.менеджер = менеджер0;
    }

    /++
    Функция загрузки в контейнер одного файла с ресурсом. 


    Returns:
        true, если файл удачно загрузился, false в противном случае.
    +/
    override protected bool загрузить_один(string имя) {
        string имя_файла_инфопанели;
        Инфопанель новая_инфопанель;
        JSONValue json_значение;
        try 
            имя_файла_инфопанели = имена_ресурсов.получить_имя_файла_ресурса(имя, ТИП_РЕСУРСА);
        catch (РесурсыИсключение) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Отсутствует файл инфопанели %s", имя));
            return false;
        }
        string текст = readText(имя_файла_инфопанели);
        try 
            json_значение = parseJSON(текст);
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Ошибка в файле инфопанели %s", имя_файла_инфопанели));
            return false;
        }

        новая_инфопанель = Инфопанель.из_JSON(json_значение, менеджер);
        if (новая_инфопанель) {
            ресурсы[имя] = новая_инфопанель;
            return true;
        }
        else {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Ошибка в файле инфопанели %s", имя_файла_инфопанели));
            return false;
        }
    }
    
}    
