/++
Модуль, представляющий класс $(LINK2 #Лампа, Лампа) - 
3D-объект, освещающий меши сцены.

Одновременно в сцене может быть до 8-ми ламп.
+/
module ddd.lamp;

import std.string;
//import derelict.opengl.gl;
import ddd.matrix, ddd.vector, ddd.dvigimyi_object, ddd.zavisimost.graphika_gl;
import my_utils.journal;

private enum имя_модуля = "лампа";

/// Класс объекта, определяющего освещение мешей сцены.
public class Лампа : ДвижимыйОбъект {

    //public static const int[] ЛАМПЫ = GL_ЛАМПЫ;

    private size_t номер_лампы;
    private float[4] окружающий_свет = [ 0.2f, 0.2f, 0.2f, 1.0f ];
    private float[4] диффузный_свет = [ 1.0f, 1.0f, 1.0f, 1.0f ];
    private float[4] бликовый_свет = [ 1.0f, 1.0f, 1.0f, 1.0f ];
    private float[4] позиция_лампы = [ 0.0f, 0.0f, 0.0f, 1.0f ];
    
    private bool изменился_окружающий_свет = true;
    private bool изменился_диффузный_свет = true;
    private bool изменился_бликовый_свет = true;
    
    private bool включено = true;

    /++
    Конструктор, принимающий имя и номер лампы в системе OpenGL.

    Номер лампы должен быть уникальным в приложении числом от 0 до 7.
    +/ 
	public this(string имя, int номер_лампы0) {
		super(имя);
		this.номер_лампы = номер_лампы0;
	}

    /++
    Функция, задающая окружающее (ambient) освещение.
    
    Params:
        свет0 = массив из 4-х компонент цвета окружающего освещения
    +/
	public void задать_окружающий_свет(const(float[4]) свет0){
		окружающий_свет = свет0;
		изменился_окружающий_свет = true;
	}

    /++
    Функция, задающая рассеянное (diffuse) освещение.
    
    Params:
        свет0 = массив из 4-х компонент цвета диффузного освещения
    +/
	public void задать_диффузный_свет(const(float[4]) свет0){
		диффузный_свет = свет0;
		изменился_диффузный_свет = true;
	}

    /++
    Функция, задающая бликовое (specular) освещение.
    
    Params:
        свет0 = массив из 4-х компонент цвета бликового освещения
    +/
	public void задать_бликовый_свет(const(float[4]) свет0){
		бликовый_свет = свет0;
		изменился_бликовый_свет = true;
	}
	
    /++
    Функция, включающая/выключающая эту $(B Лампу) в сцене.
    
    Params:
        включено0 = включить/выключить эту $(B Лампу) в сцене.
    +/
    public void включенность(bool включено0){
        включено = включено0;
    }

    /// Возвращает текущее значение включенности в сцену.
    public bool включенность(){
        return включено;
    }

    /// Меняет текущее значение включенности в сцену на противоположное.
    public void поменять_включенность(){
        включено = ! включено;
    }

    /++
    Функция, освещающая 3D-объекты этой $(B Лампой).


    Если эта $(B Лампа) отключена, то функция не выполняет никаких действий.
    +/
	override public void рисовать(){
        if (включено){
             //glEnable(GL_ЛАМПЫ[номер_лампы]); 
             gl_включить_лампу(номер_лампы);
             Матрица4 матрица = родитель.получить_унаследованную_матрицу();
             debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("Матрица лампы %s", матрица.toString()));
             Вектор3 позиция = матрица.Позиция();
             позиция_лампы[0] = позиция.x;
             позиция_лампы[1] = позиция.y;
             позиция_лампы[2] = позиция.z;
             //glLightfv(GLLIGHTS[номер_лампы], GL_POSITION, &lightPosition[0]);
             gl_задать_позицию_лампы(номер_лампы, &(позиция_лампы[0]));
             if (изменился_окружающий_свет){
                 //glLightfv(GLLIGHTS[номер_лампы], GL_AMBIENT, &окружающий_свет[0]);
                 gl_изменить_свет(номер_лампы, GL_ОКРУЖАЮЩИЙ_СВЕТ, &окружающий_свет[0]);
                 изменился_окружающий_свет = false;
             }
             if (изменился_диффузный_свет){
                 //glLightfv(GLLIGHTS[номер_лампы], GL_DIFFUSE, &диффузный_свет[0]);
                 gl_изменить_свет(номер_лампы, GL_ДИФФУЗНЫЙ_СВЕТ, &диффузный_свет[0]);
                 изменился_диффузный_свет=false;
             }
             if (изменился_бликовый_свет){
                 //glLightfv(GLLIGHTS[номер_лампы], GL_SPECULAR, &бликовый_свет[0]);
                 gl_изменить_свет(номер_лампы, GL_БЛИКОВЫЙ_СВЕТ, &бликовый_свет[0]);
                 изменился_бликовый_свет=false;
             }
             debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("Включена лампа с позицией %s", позиция_лампы));
        }
        else
             gl_выключить_лампу(номер_лампы); 
		
	}

}
