/++
Модуль, определяющий класс $(LINK2 #Материал,Материал), представляющий материал 3D-объекта в сцене.
+/
module ddd.material;

import std.string;
//import derelict.opengl3.gl;
import ddd.texture, ddd.zavisimost.graphika_gl;
import my_utils.journal;

private enum имя_модуля = "материал";

/++
Класс, представляющий материал 3D-объекта в сцене. 

Содержит значения диффузного, бликового и окружающего (однотонного) цвета материала.
Позволяет задать диффузную текстуру, в этом случае значение диффузного цвета перестаёт действовать.
+/
class Материал {
private:
    
    float[4] диффузный_цвет = [0.5f, 0.5f, 0.5f, 1.0f]; // Серый цвет
    float[4] окружающий_цвет = [0.1f, 0.1f, 0.1f, 1.0f]; // Тёмно-серый цвет
    float[4] бликовый_цвет = [0.0f, 0.0f, 0.0f, 1.0f]; // Отсутствует
    float степень_блика = 5.0f;
    
    bool задан_окружающий_цвет = false;
    
    Текстура диффузная_текстура = null;
    Текстура текстура_нормалей = null;
    Текстура бликовая_текстура = null;
    
        
public:

    /++
    Конструктор по-умолчанию.

    Создаёт материал нейтрально-серого цвета без бликов и текстур.
    +/
    this(){};

    /++
    Функция, позволяющая задать материалу диффузную текстуру.

    Params:
        текстура = $(LINK2 texture.html#Текстура, Текстура), которая будет 
            использоваться в качестве диффузной.
    +/
    void задать_диффузную_текстуру(Текстура текстура) {
        диффузная_текстура=текстура;
    }
    
    void задать_текстуру_нормалей(Текстура текстура) {
        текстура_нормалей=текстура;
    }
    
    void задать_бликовую_текстуру(Текстура текстура) {
        бликовая_текстура=текстура;
    }
    
    /++
    Функция, позволяющая задать материалу диффузный цвет.

    В случае, если не задан окружающий цвет, задаёт также его со значением в 5 раз темнее, 
    чем переданный диффузный цвет. Заданный диффузный цвет не влияет на отображение объекта,
    если была задана диффузная текстура.


    Params:
        цвет = массив из 4-х значений, который будет использоваться в качестве диффузного цвета.
    +/
    void задать_диффузный_цвет(const float[4] цвет) {
        диффузный_цвет = цвет;
        if (!задан_окружающий_цвет) {
            окружающий_цвет[] = цвет[]/5.0f ;
            окружающий_цвет[3] = 1.0f;
        }
        
    }

    /++
    Функция, позволяющая задать материалу окружающий цвет.


    Params:
        цвет = массив из 4-х значений, который будет использоваться в качестве окружающего цвета.
    +/
    void задать_окружающий_цвет(const float[4] цвет) {
        окружающий_цвет = цвет;
        задан_окружающий_цвет = true;
    }


    /++
    Функция, позволяющая задать материалу бликовый цвет.


    Params:
        цвет = массив из 4-х значений, который будет использоваться в качестве бликового цвета.
    +/
    void задать_бликовый_цвет(const float[4] цвет) {
        бликовый_цвет = цвет;
    }


    /++
    Функция, позволяющая задать материалу степень блика, т.е. параметр, который определяет размер и вид блика.


    Params:
        степень = число, которое будет использоваться в качестве степени блика.
    +/
    void задать_степень_блика(const float степень) {
        степень_блика = степень;
    }

    /++
    Вывод параметров этого материала в систему OpenGL. Применяется непосредственно перед отрисовкой 
    меша 3D-объекта, который использует этот материал.
    +/
    void рисовать() {

        gl_задать_окружающий_материал(окружающий_цвет);
        //glMaterialfv(GL_FRONT, GL_AMBIENT, &окружающий_цвет[0]);
        
        if (бликовая_текстура is null ) {
            gl_задать_бликовый_материал(бликовый_цвет);
            //glMaterialfv(GL_FRONT, GL_SPECULAR, &бликовый_цвет[0]);
        }
        
        if (диффузная_текстура is null ) {
            gl_отключить_текстуры();
            gl_задать_диффузный_материал(диффузный_цвет);
            //glMaterialfv(GL_FRONT, GL_DIFFUSE, &диффузный_цвет[0]);
        }
        else {
           диффузная_текстура.рисовать();
        }

        gl_задать_степень_блика(степень_блика);
        //glMaterialf(GL_FRONT, GL_SHININESS, степень_блика);
        //if (!is null текстура_нормалей) {
       //     
        //}
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, this.toString());
        
    }

    /// Вывод параметров этого материала в строковое представление.
    override string toString() {
        string результат = "Материал: ";
        результат ~= format("Диффуз(%(%.2g, %)) ", диффузный_цвет);
        результат ~= format("Окружающий(%(%.2g, %)) ", окружающий_цвет);
        результат ~= format("Блик(%(%.2g, %)) ", бликовый_цвет);
        результат ~= format("Диффузная текстура(%s) ", &диффузная_текстура);

        
        return результат;
    }
    
}
