/++
Модуль, определяющий классы 3D-объектов (мешей), которые возможно вывести на экран:
$(UL  $(LI  $(LINK2 #Меш, Меш) - для статических 3D-объектов)
$(LI  $(LINK2 #МешСоСкелетом, МешСоСкелетом) - для 3D-объектов, к которым возможно применять скелетную анимацию)
)

$(BR)
Дополнительно определены:
$(UL  
$(LI  класс $(LINK2 #МешИсключение, МешИсключение) - для исключений, которые могут возникнуть при работе с мешами)
$(LI  функция $(LINK2 #выбор_меша, выбор_меша) - помогающая выяснить, 
к какому именно классу относится тот или иной файл с ресурсом меша)
$(LI  класс $(LINK2 #ЗагруженныеМеши, ЗагруженныеМеши), позволяющий получать объекты мешей
из файлов-ресурсов с расширением ".me". 
Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты класса $(LINK2 #Меш, Меш) 
(или, возможно, $(LINK2 #МешСоСкелетом, МешСоСкелетом)) через него.)
)
+/
module ddd.mesh;

//import derelict.opengl.gl;
import std.file, std.array, std.string;
import my_utils.journal;
import ddd.resources, ddd.zavisimost.graphika_gl, ddd.vector, ddd.skelet;

import std.stdio;


private enum имя_модуля = "меш";
private enum ТИП_РЕСУРСА = "меш";

private enum РАЗМЕР_ВЕКТОР3_FLOAT = 3*4;
private enum РАЗМЕР_ВЕКТОР2_FLOAT = 2*4;
private enum РАЗМЕР_ВЕКТОР3_SHORT = 3*2;
private enum РАЗМЕР_SHORT = 2;

private enum РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ = 6;

/// Вариант отрисовки меша, в котором буферы вершин, текстур и нормалей слиты в один чередующийся буфер.
enum ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ = false;

/// Вариант отрисовки меша, в котором буферы вершин, текстур и нормалей отдельны.
enum ОТДЕЛЬНЫЕ_БУФЕРЫ = true;

private enum byte[] ЗАГОЛОВОК_МЕША_1_ВЕРСИИ = [68, 68, 68, 32, 1, 0]; //, 0, 0];
private enum byte[] ЗАГОЛОВОК_МЕША_2_ВЕРСИИ = [68, 68, 68, 32, 2, 0]; //, 0, 0];

/++
Класс исключений, которые могут возникнуть при работе с мешами.
+/
class МешИсключение : Exception {
    this(string s) { 
        super(s); 
    }
};

/++
Класс, представляющий геометрические данные о статическом 3D-объекте.
+/
class Меш {
    protected void* буфер_вершин;
    protected void* буфер_текстур;
    protected void* буфер_индексов;
    protected void* буфер_нормалей;
    
    protected byte[] основной_буфер;
    protected float[] чередующийся_буфер_F;
    protected void*  чередующийся_буфер;
    
    
    protected ushort _граней, _вершин;
    private bool тип_отрисовки; 

    protected this() {}

    /++
    Конструктор, принимающий имя файла с данными о меше.

    Params:
        имя_файла_меша = Имя файла типа ".me" с данными меша формата первой версии.
        тип_отрисовки0 = В каком виде будут представлены данные о координатах вершин, 
            текстур и нормалей внутри класса, см. $(LINK2 #ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ, ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ) 
            и $(LINK2 #ОТДЕЛЬНЫЕ_БУФЕРЫ, ОТДЕЛЬНЫЕ_БУФЕРЫ)

    Throws:
        $(LINK2 #МешИсключение, МешИсключение), если заголовок файла не соответствует заголовку файла меша первой версии.
    +/
    this(string имя_файла_меша, bool тип_отрисовки0=ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ) {
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("Начинаем загрузку меша %s", имя_файла_меша));
        enum РАЗМЕР_ЗАГОЛОВКА = 10;
        auto байты1 = read(имя_файла_меша);
        
        //writeln("Начинаем загрузку меша", имя_файла_меша);
        if (байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ] != ЗАГОЛОВОК_МЕША_1_ВЕРСИИ) {
            writeln(байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ]);
            writeln("Случилась ошибка заголовка при чтении файла меша");
            throw new МешИсключение("Заголовок файла не соответствует требуемому для меша 1 версии!"); 
        }
        ushort[] длины = cast(ushort[]) байты1[РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ..РАЗМЕР_ЗАГОЛОВКА];
        _вершин = длины[0];
        _граней = длины[1];
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format(":байты1=%s", &байты1[0]));
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("вершин=%s,  граней=%s", _вершин, _граней));
        int размер_буфера = _вершин * (2 * РАЗМЕР_ВЕКТОР3_FLOAT + РАЗМЕР_ВЕКТОР2_FLOAT) + _граней * РАЗМЕР_ВЕКТОР3_SHORT;
        основной_буфер = cast(byte[]) байты1[РАЗМЕР_ЗАГОЛОВКА .. (РАЗМЕР_ЗАГОЛОВКА+размер_буфера)];
        тип_отрисовки = тип_отрисовки0;
        _присвоить_буферы();
        
        в_журнал(имя_модуля, УровниЖурнала.Информация, "Меш, типа, загружен");
    }

    /++
    Конструктор, принимающий координаты вершин, текстур и нормалей, а также индексы для построения граней 
    в виде массивов.

    Params:
        координаты_вершин = массив с координатами вершин (3 координаты на вершину)
        координаты_текстур = массив с текстурными координатами (2 координаты на вершину)
        координаты_нормалей = массив с координатами нормалей (3 координаты на вершину)
        индексы = массив индексов вершин, определяющих грани (3 индекса на грань)
        тип_отрисовки0 = В каком виде будут представлены данные о координатах вершин, 
            текстур и нормалей внутри класса, см. $(LINK2 #ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ, ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ) 
            и $(LINK2 #ОТДЕЛЬНЫЕ_БУФЕРЫ, ОТДЕЛЬНЫЕ_БУФЕРЫ)
    +/
    this(float[] координаты_вершин, float[] координаты_текстур, float[] координаты_нормалей, 
         ushort[] индексы, bool тип_отрисовки0=ЧЕРЕДУЮЩИЕСЯ_БУФЕРЫ) {
         в_журнал(имя_модуля, УровниЖурнала.Информация, "Начинаем формирование меша.");
        _вершин = cast(ushort) координаты_вершин.length/3;
        _граней = cast(ushort) индексы.length/3;
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("_вершин=%s  _граней=%s", _вершин,_граней));
        byte[] вершины_B = cast(byte[]) координаты_вершин;
        byte[] текстуры_B = cast(byte[]) координаты_текстур;
        byte[] нормали_B = (координаты_нормалей is null) ? new byte[_вершин*РАЗМЕР_ВЕКТОР3_FLOAT] : cast(byte[]) координаты_нормалей;
        //byte[] normKbyte=(координаты_нормалей is null) ? new byte[_вершин*РАЗМЕР_ВЕКТОР3_FLOAT] : cast(byte[]) координаты_нормалей;
        byte[] индексы_B = cast(byte[]) индексы;
        основной_буфер = вершины_B ~ текстуры_B ~ нормали_B ~ индексы_B;
        тип_отрисовки = тип_отрисовки0;
        _присвоить_буферы();
        в_журнал(имя_модуля, УровниЖурнала.Информация, "Меш, типа, сформирован.");
        
    }

    private void _присвоить_буферы() {
        if (тип_отрисовки)
            _присвоить_отдельные_буферы();
        else
            _присвоить_чередующийся_буфер();
    }
    
    protected void _присвоить_отдельные_буферы() {
        int vB0 = 0;
        int tB0 = vB0+_вершин*РАЗМЕР_ВЕКТОР3_FLOAT;
        int nB0 = tB0+_вершин*РАЗМЕР_ВЕКТОР2_FLOAT;
        int ib0 = nB0+_вершин*РАЗМЕР_ВЕКТОР3_FLOAT;
        int iB1 = ib0+_граней*РАЗМЕР_ВЕКТОР3_SHORT;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("vB0=%s    tB0=%s    nB0=%s    ib0=%s    iB1=%s", vB0,  tB0,  nB0, ib0, iB1));
        буфер_вершин =  cast(void*) основной_буфер[vB0..tB0];
        буфер_текстур = cast(void*) основной_буфер[tB0..nB0];
        буфер_нормалей =  cast(void*) основной_буфер[nB0..ib0];
        буфер_индексов =   cast(void*) основной_буфер[ib0..iB1];
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format(":буфер_вершин=%s", буфер_вершин));
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format(":буфер_текстур=%s", буфер_текстур));
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format(":буфер_нормалей=%s", буфер_нормалей));
        
    }

    protected void _присвоить_чередующийся_буфер() {
        float[] interBuf=new float[_вершин*8];
        int vB0 = 0;
        int tB0 = vB0 + _вершин * РАЗМЕР_ВЕКТОР3_FLOAT;
        int nB0 = tB0 + _вершин * РАЗМЕР_ВЕКТОР2_FLOAT;
        int ib0 = nB0 + _вершин * РАЗМЕР_ВЕКТОР3_FLOAT;
        int iB1 = ib0 + _граней * РАЗМЕР_ВЕКТОР3_SHORT;
        for (int i=0; i<_вершин; i++) {
            float[] tex = cast(float[]) основной_буфер[tB0+i*РАЗМЕР_ВЕКТОР2_FLOAT .. tB0+i*РАЗМЕР_ВЕКТОР2_FLOAT+РАЗМЕР_ВЕКТОР2_FLOAT];
            float[] norm = cast(float[]) основной_буфер[nB0+i*РАЗМЕР_ВЕКТОР3_FLOAT .. nB0+i*РАЗМЕР_ВЕКТОР3_FLOAT+2*РАЗМЕР_ВЕКТОР3_FLOAT];
            float[] vert = cast(float[]) основной_буфер[vB0+i*РАЗМЕР_ВЕКТОР3_FLOAT .. vB0+i*РАЗМЕР_ВЕКТОР3_FLOAT+2*РАЗМЕР_ВЕКТОР3_FLOAT];
            interBuf[i*8+0] = tex[0];  interBuf[i*8+1] = tex[1]; 
            interBuf[i*8+2] = norm[0]; interBuf[i*8+3] = norm[1]; interBuf[i*8+4]=norm[2]; 
            interBuf[i*8+5] = vert[0]; interBuf[i*8+6] = vert[1]; interBuf[i*8+7]=vert[2]; 
        }
        буфер_индексов = cast(void*) основной_буфер[ib0..iB1];
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("interBuf=%s", interBuf[0..24])); 
        чередующийся_буфер_F = interBuf;
        чередующийся_буфер = cast(void*)(cast(byte[])( interBuf));
    }

    /++
    Вывод 3D-объекта этого меша на экран.
    +/
    void рисовать() {
        if (тип_отрисовки)
            gl_рисовать_отдельные_буферы(_граней, буфер_вершин, буфер_нормалей, буфер_текстур, буфер_индексов);
        else
            gl_рисовать_чередующийся_буфер(_граней, чередующийся_буфер, буфер_индексов);
    }

    /// Возвращает количество граней в этом меше.
    public int количество_граней(){
          return _граней;
    }

    /// Деструктор, очищающий все буферы.
    ~this() {
        буфер_вершин = null;
        буфер_текстур = null;
        буфер_нормалей = null;
        буфер_индексов = null;
        чередующийся_буфер = null;
    }

    /// Функция для совместимости с анимированным мешем. В этом классе не выполняет никаких действий.
    void присвоить_скелет(Скелет скелет0) {}

    /++
    Функция для совместимости с анимированным мешем. В этом классе не выполняет никаких действий.

    Returns:
        false
    +/
    bool установить_состояние_анимации(string название_анимации, float положение_кадра) {
        return false;
    }

    /++
    Функция для совместимости с анимированным мешем. В этом классе не выполняет никаких действий.


    Returns:
        false
    +/
    bool объединить_состояния_анимации(string название_анимации1, float положение_кадра1,
                                       string название_анимации2, float положение_кадра2, float коэффициент) {
        return false;
    }

    /++
    Вывод части данных этого меша в строковое представление. 

    Осторожно, может быть очень большая строка!
    +/
    override string toString() {
        int длина_массива = _вершин*РАЗМЕР_ВЕКТОР3_FLOAT;
        float[] вершины = cast(float[]) основной_буфер[0 .. длина_массива];
        return "Меш(\n" ~ format("вершин=%s,  граней=%s,\n", _вершин, _граней) ~
            format("положение вершин=%s,\n", вершины);
    }

    /++
    Функция, заменяющая текущие текстурные координаты на новые. Для текстурной анимации.

    Note:
        При применении этой функции тип отрисовки этого меша переводится на отдельные буферы.
    +/
    public void изменить_координаты_текстур(float[] координаты_текстур) {
        if (!тип_отрисовки) {
            чередующийся_буфер_F.length = 0;
            чередующийся_буфер = null;
            тип_отрисовки = ОТДЕЛЬНЫЕ_БУФЕРЫ;
            _присвоить_буферы();
        }
        byte[] байтовые_координаты_текстур = cast(byte[]) координаты_текстур;
        int tB0 = _вершин * РАЗМЕР_ВЕКТОР3_FLOAT;
        int nB0 = tB0 + _вершин * РАЗМЕР_ВЕКТОР2_FLOAT;
        //writeln("&основной_буфер1=", &основной_буфер);
        основной_буфер[tB0..nB0] = байтовые_координаты_текстур;
        //writeln("&основной_буфер2=", &основной_буфер);
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("Меш: изменён буфер текстур=%s", буфер_текстур));
    }
    
}

/++
Класс, представляющий геометрические данные о 3D-объекте, к которому возможно применять скелетную анимацию. 
Для этого ему нужно присвоить уже существующий объект $(LINK2 skelet.html#Скелет, Скелет), 
имена костей которого должны соответствовать группам вершин меша. Анимация 
будет рассчитываться совместно обоими классами.

Тип отрисовки этого класса всегда через отдельные буферы.
+/
class МешСоСкелетом : Меш {
    protected string[] имена_групп;
    protected float[int][int] веса_вершин;
    protected Вектор3[] начальное_положение_вершин, анимированные_вершины;
    protected Скелет скелет = null;

    /++
    Конструктор, принимающий имя файла с данными о меше.


    Params:
        имя_файла_меша = Имя файла типа ".me" с данными меша формата первой версии.


    Throws:
        $(LINK2 #МешИсключение, МешИсключение), если заголовок файла не соответствует заголовку файла меша второй версии.
    +/
    this(string имя_файла_меша) {
        super();
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("Начинаем загрузку меша со скелетом %s", имя_файла_меша));
        enum РАЗМЕР_ЗАГОЛОВКА = 14;
        enum РАЗМЕР_ЗАПИСИ_С_ВЕСОМ = 8;
        auto байты1 = read(имя_файла_меша);
        
        //writeln("Начинаем загрузку меша со скелетом ", имя_файла_меша);
        if (байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ] != ЗАГОЛОВОК_МЕША_2_ВЕРСИИ) {
            writeln(байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ]);
            writeln("Случилась ошибка заголовка при чтении файла меша");
            throw new МешИсключение("Заголовок файла не соответствует требуемому для меша 2 версии!"); 
        }
        ushort[] длины = cast(ushort[]) байты1[РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ..РАЗМЕР_ЗАГОЛОВКА];
        _вершин = длины[0];
        _граней = длины[1];
        ushort количество_групп = длины[2];
        ushort весов = длины[3];
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format(":байты1=%s", &байты1[0]));
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("вершин=%s,  граней=%s, групп=%s,  весов=%s",
                _вершин, _граней, количество_групп, весов));
        
        size_t размер_буфера = _вершин * (2 * РАЗМЕР_ВЕКТОР3_FLOAT + РАЗМЕР_ВЕКТОР2_FLOAT) + _граней * РАЗМЕР_ВЕКТОР3_SHORT;
        основной_буфер = cast(byte[]) байты1[РАЗМЕР_ЗАГОЛОВКА .. (РАЗМЕР_ЗАГОЛОВКА+размер_буфера)];
        _присвоить_отдельные_буферы();

        начальное_положение_вершин = uninitializedArray!(Вектор3[])(_вершин);
        анимированные_вершины = uninitializedArray!(Вектор3[])(_вершин);
        for (int i=0; i<_вершин; i++) {
            float[3] xyz = cast(float[]) буфер_вершин[i*РАЗМЕР_ВЕКТОР3_FLOAT .. (i+1)*РАЗМЕР_ВЕКТОР3_FLOAT];
            начальное_положение_вершин[i] = Вектор3(xyz);
            анимированные_вершины[i] = Вектор3(xyz);
        }
        
        size_t место_чтения = РАЗМЕР_ЗАГОЛОВКА+размер_буфера;
        ushort длина_имени_группы;
        for (int i=0; i<количество_групп; i++) {
            длина_имени_группы = (cast(ushort[]) байты1[место_чтения .. (место_чтения+РАЗМЕР_SHORT)])[0];
            место_чтения += РАЗМЕР_SHORT;
            имена_групп ~= cast(string) байты1[место_чтения .. (место_чтения+длина_имени_группы)];
            место_чтения += длина_имени_группы;
        }

        ushort индекс_группы, индекс_вершины;
        float вес;
        for (int i=0; i<весов; i++) {
            индекс_группы = (cast(ushort[]) байты1[место_чтения .. (место_чтения+РАЗМЕР_SHORT)])[0];
            индекс_вершины = (cast(ushort[]) байты1[(место_чтения+РАЗМЕР_SHORT) .. (место_чтения+РАЗМЕР_SHORT*2)])[0];
            вес = (cast(float[]) байты1[(место_чтения+РАЗМЕР_SHORT*2) .. (место_чтения+РАЗМЕР_ЗАПИСИ_С_ВЕСОМ)])[0];
            место_чтения += РАЗМЕР_ЗАПИСИ_С_ВЕСОМ;
            веса_вершин[индекс_вершины][индекс_группы] = вес;
        }
        
        
        в_журнал(имя_модуля, УровниЖурнала.Информация, "Меш, типа, загружен");
    }

    /++
    Вывод части данных этого меша в строковое представление. 


    Осторожно, может быть очень большая строка!
    +/
    override string toString() {
        return "МешСоСкелетом(\n" ~ format("вершин=%s,  граней=%s,\n", _вершин, _граней) ~
            format("имена групп=%s\n", имена_групп) ~ format("веса вершин=%s,\n", веса_вершин) ~
            format("начальное положение вершин=%s,\n", начальное_положение_вершин) ~
            format("текущее положение вершин=%s)", анимированные_вершины);
    }

    /++
    Присваивает существующий скелет этому мешу.
    +/
    override void присвоить_скелет(Скелет скелет0) {
        скелет = скелет0;
    }

    /++
    Функция, устанавливающая форму (т.е. координаты вершин) этого меша в требуемое состояние анимации.

    Если этому мешу не был присвоен скелет, то функция не выполняет никаких действий.
    
    Params:
        название_анимации = Имя анимации, должно быть одним из существуюших имён анимаций 
            (или, "действий") присвоенного скелета.
        положение_кадра = Номер кадра выбранной анимации. Возможно использовать нецелое число, 
            тогда будет вычислена интерполяция между ближайшими кадрами.

    Returns:
        true, если установка формы успешно выполнена, false в противном случае.
    +/
    override bool установить_состояние_анимации(string название_анимации, float положение_кадра) {
        if (скелет is null) 
            return false; // Начальное состояние в буфер_вершин попало в конструкторе
        auto состояние_анимации0 = скелет.получить_состояние_анимации(название_анимации, положение_кадра);
        return _установить_состояние_анимации(состояние_анимации0);
    }

    /++
    Функция, устанавливающая форму (т.е. координаты вершин) этого меша в требуемое 
    состояние объединения двух анимаций.


    Если этому мешу не был присвоен скелет, то функция не выполняет никаких действий.
    
    Params:
        название_анимации1 = Имя первой анимации, должно быть одним из существуюших имён анимаций 
            (или, "действий") присвоенного скелета.
        положение_кадра1 = Номер кадра выбранной первой анимации. Возможно использовать нецелое число, 
            тогда будет вычислена интерполяция между ближайшими кадрами.
        название_анимации2 = Имя второй анимации, должно быть одним из существуюших имён анимаций 
            (или, "действий") присвоенного скелета.
        положение_кадра2 = Номер кадра выбранной второй анимации. Возможно использовать нецелое число, 
            тогда будет вычислена интерполяция между ближайшими кадрами.
        коэффициент = Если вычисляется состояние кости скелета, которая участвует в обеих анимациях,
        то этот параметр используется как коэффициент интерполяции между ними. Если кость,
        для которой вычисляется состояние, используется только в одной из анимаций, то 
        этот параметр не используется.

    Returns:
        true, если установка формы успешно выполнена, false в противном случае.
    +/
    override bool объединить_состояния_анимации(string название_анимации1, float положение_кадра1,
                                                string название_анимации2, float положение_кадра2, 
                                                float коэффициент) {
        if (скелет is null) 
            return false; // Начальное состояние в буфер_вершин попало в конструкторе
        auto состояние_анимации0 = скелет.объединить_состояния_анимации(название_анимации1, положение_кадра1,
                                                                        название_анимации2, положение_кадра2, коэффициент);
        return _установить_состояние_анимации(состояние_анимации0);
    }
    
    private bool _установить_состояние_анимации(СостояниеАнимацииКости[string] состояние_анимации0) {
        СостояниеАнимацииКости[] состояние_анимации1;
        foreach (кость; имена_групп) {
            if (кость in состояние_анимации0) 
                состояние_анимации1 ~= состояние_анимации0[кость];
            else
                состояние_анимации1 ~= СОСТОЯНИЕ_БЕЗ_ИЗМЕНЕНИЙ;
        }
        for (int i=0; i<_вершин; i++) {
            Вектор3 сдвиг_вершины = Вектор3(Вектор3.НОЛЬ);
            float общий_вес = 0;
            foreach (номер_кости, вес; веса_вершин[i]) {
                общий_вес += вес;
                auto изменения_кости = состояние_анимации1[номер_кости];
                if (изменения_кости.типы_изменения == 0) 
                    continue;
                Вектор3 текущий_вектор = начальное_положение_вершин[i] - изменения_кости.начальное_состояние.положение;
                //auto изменения_кости = состояние_анимации1[номер_кости];
                if (изменения_кости.типы_изменения & ТипыИзменения.ориентация)
                    текущий_вектор = изменения_кости.изменение_ориентации * текущий_вектор;
                if (изменения_кости.типы_изменения & ТипыИзменения.масштаб)
                    текущий_вектор *= изменения_кости.изменение_масштаба;
                if (изменения_кости.типы_изменения & ТипыИзменения.положение)
                    текущий_вектор += изменения_кости.изменение_положения;
                текущий_вектор += изменения_кости.начальное_состояние.положение - начальное_положение_вершин[i];
                debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("группа=%s, вес=%s, текущий_вектор=%s", 
                                                                         имена_групп[номер_кости],
                                                                         вес,
                                                                         текущий_вектор
                                                                        ));
                сдвиг_вершины += текущий_вектор * вес;
            }
            анимированные_вершины[i] = начальное_положение_вершин[i] + (сдвиг_вершины * (1.0/общий_вес));
        }
        буфер_вершин =  cast(void*) анимированные_вершины;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("анимированные_вершины=%s", анимированные_вершины));
        return true;        
    }

    
    /++
    Вывод 3D-объекта этого меша на экран.
    +/
    override void рисовать() {
        gl_рисовать_отдельные_буферы(_граней, буфер_вершин, буфер_нормалей, буфер_текстур, буфер_индексов);
    }
}    

/++
Функция, строящая объект того класса меша, который указан в заголовке переданного файла.

Params:
    имя_файла_меша = Имя (с путём к нему) файла, для которого определяется тип меша.

Returns:
    Объект одного из классов: $(LINK2 #Меш, Меш) или $(LINK2 #МешСоСкелетом, МешСоСкелетом), в зависимости 
    от заголовка переданного файла.

Throws:
    $(LINK2 #МешИсключение, МешИсключение), если заголовок файла не соответствует ни 
    одному из заголовков файла меша первой или второй версии.
+/
Меш выбор_меша(string имя_файла_меша) {
    auto байты1 = read(имя_файла_меша, РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ);
    if (байты1 == ЗАГОЛОВОК_МЕША_1_ВЕРСИИ) 
        return new Меш(имя_файла_меша);
    if (байты1 == ЗАГОЛОВОК_МЕША_2_ВЕРСИИ) 
        return new МешСоСкелетом(имя_файла_меша);
    throw new МешИсключение("Заголовок файла не соответствует требуемому для меша 1 или 2 версии!"); 
}

/++
Класс-контейнер ресурсов объектов класса $(LINK2 #Меш, Меш) и его потомков.


Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты классов мешей, построенных из файлов, через него.
+/
class ЗагруженныеМеши : ЗагруженныеРесурсы!(Меш) {
    
    /++
    Конструктор, принимающий ссылку на объект типа $(LINK2 resources.html#ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов.
    +/
    this(ИменаРесурсов имена_ресурсов0) {
        super(имена_ресурсов0);
    }
    
    /++
    Функция загрузки в контейнер одного файла с ресурсом. 


    Returns:
        true, если файл удачно загрузился, false в противном случае.
    +/
    override protected bool загрузить_один(string имя) {
        string имя_файла_меша;
        Меш новый_меш;
        try 
            имя_файла_меша = имена_ресурсов.получить_имя_файла_ресурса(имя, ТИП_РЕСУРСА);
        catch (РесурсыИсключение) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Отсутствует файл меша %s", имя));
            return false;
        }
        try 
            новый_меш = выбор_меша(имя_файла_меша);
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Ошибка в файле меша %s", имя_файла_меша));
            return false;
        }
        ресурсы[имя] = новый_меш;
        return true;
    }
    

    
    
}
