/++
Модуль, представляющий класс $(LINK2 #Место, Место) - контейнер для 3D-объектов.
Каждое $(B Место) может быть родителем/потомком другого $(B Места). В объекте класса
сохраняются данные о положении, повороте и масштабе. Все эти пространственные характеристики 
применяются ко всем содержащимся в этом контейнере объектам и ко всем объектам,
содержащимся в контейнерах-потомках этого $(B Места). Все объекты $(B Место), 
в которых содержатся видимые на экране 3D-объекты, составляют иерархию, корень которой 
(под именем $(B начальное_место)) 
содержится в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
+/
module ddd.mesto;

//import derelict.opengl.gl;
import std.file, std.string, std.conv, std.math;
import my_utils.journal;

import ddd.quaternion, ddd.vector, ddd.matrix, ddd.dvigimyi_object, ddd.zavisimost.graphika_gl;


enum имя_модуля = "место";

/// Класс-контейнер для 3D-объектов, определяющий их расположение, поворот и масштаб
public class Место {
    /// Коэффициент для перевода значения угла в радианах в значение в градусах
    public enum float РАДИАНЫ_В_ГРАДУСЫ = to!(float)(180.0/PI);
    
    private Место родитель;                     // Родительское место
    private Место[string] дети;              // Массив указателей на дочерние места
    private ДвижимыйОбъект[string] объекты;     // Массив указателей на привязанные объекты
    private string имя;                         // Имя этого места
    private Кватернион ориентация;              // Сохраняет ориентацию места относительно его родителя.
    private Вектор3 позиция;                    // Сохраняет позицию/сдвиг места относительно его родителя.
    private Вектор3 масштаб;                    // Сохраняет коэффициент масштабирования, применяемый к этому месту
    
    private bool включено = true;               // Если не включено, то все объекты этого места и его детей не выводятся на экран
    
    private Матрица4 кэшированное_преобразование;
    private bool кэшированное_преобразование_устарело = true;

    private static Место[string] ПУСТОЙ_МАССИВ_ДЕТЕЙ;
    private static ДвижимыйОбъект[string] ПУСТОЙ_МАССИВ_ОБЪЕКТОВ;

    /++
    Конструктор, создающий объект $(B Места) с расположением в начале координат, без поворота и с 
    единичным масштабом.

    Params:
        имя0 = имя, которое будет присвоено создаваемому $(B Месту)
    +/
    public this(string имя0){
        имя = имя0;
        позиция = Вектор3.НОЛЬ;
        масштаб = Вектор3.ЕДИНИЦА_МАСШТАБА;
        ориентация = Кватернион.НОЛЬ;

        родитель = null;
        //кэшированное_преобразование = Matrix4();
    }

    /// Сделать этот объект потомком другого $(B Места)
    public void задать_родителя(Место родительское_место){
        родитель = родительское_место;
    }

    /// Метод, возвращающий место, являющееся родетелем этого $(B Места)
    public @property Место получить_родителя(){
        return родитель;
    }

    /// Метод, возвращающий имя этого $(B Места)
    public @property string получить_имя() {
        return имя;
    }

    /// Метод, возвращающий кватернион поворота этого $(B Места)
    public @property Кватернион получить_ориентацию(){
        return ориентация;
    }

    /// Метод, возвращающий вектор расположения этого $(B Места)
    public @property Вектор3 получить_позицию(){
        return позиция;
    }

    /// Метод, возвращающий вектор масштаба этого $(B Места)
    public @property Вектор3 получить_масштаб(){
        return масштаб;
    }

    /++
    Метод, задающий кватернион поворота этого $(B Места)

    Params:
        к = кватернион, который будет задавать поворот этого $(B Места)
    +/
    public void задать_ориентацию(const(Кватернион) к){
        ориентация = Кватернион(к);
        ориентация.нормализовать();
        кэшированное_преобразование_устарело = true;
    }

    /// Метод, приводящий ориентацию этого $(B Места) в значение нулевого поворота
    public void сбросить_ориентацию(){
        ориентация = Кватернион.НОЛЬ;
        кэшированное_преобразование_устарело = true;
    }

    /++
    Метод, задающий позицию этого $(B Места)

    Params:
        позиция0 = вектор, задающий позицию этого $(B Места)
    +/
    public void задать_позицию(const(Вектор3) позиция0){
        позиция = Вектор3(позиция0);
        кэшированное_преобразование_устарело = true;
    }

    /++
    Метод, задающий позицию этого $(B Места)

    Params:
        x = координата $(B x) результирущей позиции этого $(B Места)
        y = координата $(B y) результирущей позиции этого $(B Места)
        z = координата $(B z) результирущей позиции этого $(B Места)
    +/
    public void задать_позицию(float x, float y, float z){
        позиция = Вектор3(x, y, z);
        кэшированное_преобразование_устарело = true;
    }
    
    /++
    Метод, задающий масштаб этого $(B Места)

    Params:
        масштаб0 = вектор, задающий масштаб этого $(B Места)
    +/
    public void задать_масштаб(const(Вектор3) масштаб0){
        масштаб = Вектор3(масштаб0);
        кэшированное_преобразование_устарело = true;
    }

    /++
    Метод, задающий масштаб этого $(B Места)

    Params:
        x = масштаб этого $(B Места) по координате $(B x)
        y = масштаб этого $(B Места) по координате $(B y)
        z = масштаб этого $(B Места) по координате $(B z)
    +/
    public void задать_масштаб(float x, float y, float z){
        масштаб = Вектор3(x,y,z);
        кэшированное_преобразование_устарело = true;
    }

    /++
    Функция масштабирования этого $(B Места)

    Params:
        изменение_масштаба = вектор масштабирования, в соответствии с которым изменяется масштаб этого $(B Места)
    +/
    public void масштабировать(const(Вектор3) изменение_масштаба){
        масштаб *= изменение_масштаба;
        кэшированное_преобразование_устарело = true;
    }

    /++
    Функция масштабирования этого $(B Места)

    Params:
        x = координата $(B x) вектора масштабирования, в соответствии с которым изменяется масштаб этого $(B Места)
        y = координата $(B y) вектора масштабирования, в соответствии с которым изменяется масштаб этого $(B Места)
        z = координата $(B z) вектора масштабирования, в соответствии с которым изменяется масштаб этого $(B Места)
    +/
    public void масштабировать(float x, float y, float z){
        масштаб = Вектор3(масштаб.x*x, масштаб.y*y, масштаб.z*z);
        кэшированное_преобразование_устарело = true;
    }
    
    /++
    Функция сдвига этого $(B Места)

    Params:
        смещение = вектор, на который сдвигается это $(B Место)
    +/
    public void сдвинуть(const(Вектор3) смещение){
        позиция += смещение;
        кэшированное_преобразование_устарело = true;
    }
    
    /++
    Функция сдвига этого $(B Места)

    Params:
        x = расстояние по координате $(B x), на которое сдвигается это $(B Место)
        y = расстояние по координате $(B y), на которое сдвигается это $(B Место)
        z = расстояние по координате $(B z), на которое сдвигается это $(B Место)
    +/
    public void сдвинуть(float x, float y, float z){
        позиция = Вектор3(позиция.x+x, позиция.y+y, позиция.z+z);
        кэшированное_преобразование_устарело = true;
    }
    
    /++
    Функция поворота этого $(B Места) вокруг оси $(B Z)

    Params:
        угол = угол, на который поворачивается это $(B Место)
    +/
    public void крен(float угол) //Roll
    {
        вращать(Вектор3.ЕДИНИЦА_Z, угол);
    }

    /++
    Функция поворота этого $(B Места) вокруг оси $(B X)

    Params:
        угол = угол, на который поворачивается это $(B Место)
    +/
    public void наклон(float угол) //Pitch
    {
        вращать(Вектор3.ЕДИНИЦА_X, угол);
    }

    /++
    Функция поворота этого $(B Места) вокруг оси $(B Y)

    Params:
        угол = угол, на который поворачивается это $(B Место)
    +/
    public void поворот(float угол) //Yaw
    {
        вращать(Вектор3.ЕДИНИЦА_Y, угол);
    }
    
    /++
    Функция поворота этого $(B Места) вокруг заданной оси

    Params:
        ось = вектор, представляющий ось, вокруг которой поворачивается это $(B Место)
        угол = угол, на который поворачивается это $(B Место)
    +/
    public void вращать(const(Вектор3) ось, float угол)
    {
        Кватернион к;
        к.из_угла_и_оси(угол, ось);
        вращать(к);
    }
    
    /++
    Функция поворота этого $(B Места)

    Params:
        к = кватернион, который представляет поворот, изменяющий ориентацию этого $(B Места)
    +/
    public void вращать(const(Кватернион) к){
        Кватернион к_норм = Кватернион(к);
        к_норм.нормализовать();
        ориентация *= к_норм;
        //ориентация = ориентация * к_норм;
        кэшированное_преобразование_устарело = true;
    }

    /++
    Функция, создающая новый объект $(B Места), который становится потомком этого $(B Места)

    Params:
        имя0 = имя нового $(B Места)
        сдвиг = вектор, задающий позицию нового $(B Места) относительно расположения этого $(B Места)
        вращение = кватернион, который представляет ориентацию нового $(B Места) относительно ориентации этого $(B Места)
        масштабирование = вектор, задающий масштаб нового $(B Места)
    +/
    public Место создать_ребёнка(string имя0, const(Вектор3) сдвиг, const(Кватернион) вращение, 
                                 const(Вектор3) масштабирование=Вектор3.ЕДИНИЦА_МАСШТАБА){
        Место новое_место = new Место(имя0);
        новое_место.сдвинуть(сдвиг);
        новое_место.вращать(вращение);
        добавить_ребёнка(новое_место);
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("Создано дочернее место %s", новое_место.toString()));
        return новое_место;
    }
    
    /++
    Функция, добавляющая переданный объект $(B Места) в список потомков этого $(B Места).

    Если объект ребёнок уже имеет родителя, то функция не выполняет никаких действий.


    Params:
        ребёнок = объект $(B Места), который будет потомком этого $(B Места)
    +/
    public void добавить_ребёнка(Место ребёнок){
        if (ребёнок.родитель !is null) return;
        дети[ребёнок.получить_имя] = ребёнок;
        ребёнок.задать_родителя(this);
    }
    
    /++
    Функция, возвращающая количество потомков этого $(B Места).
    +/
    public short количество_детей(){
        return to!(short) (дети.length);
    }
    
    /++
    Функция, возвращающая объект-ребёнок этого $(B Места) по его имени.
    
    Params:
        имя0 = имя искомого объекта
    +/
    public Место получить_ребёнка(string имя0){
        return дети[имя0];
    }
    
    /++
    Функция, удаляющая переданный объект-ребёнок из списка потомков этого $(B Места).
    
    Params:
        ребёнок = объект $(B Места), который будет удалён из списка потомков этого $(B Места)


    Returns:
        Объект $(B Места), удалённый из списка потомков.
    +/
    public Место удалить_ребёнка(Место ребёнок){
        return удалить_ребёнка(ребёнок.получить_имя);
    }
    
    /++
    Функция, удаляющая объект-ребёнок из списка потомков этого $(B Места).
    
    Params:
        имя_ребёнка = имя объекта $(B Места), который будет удалён из списка потомков этого $(B Места)

    Returns:
        Объект $(B Места), удалённый из списка потомков.
    +/
    public Место удалить_ребёнка(string имя_ребёнка){
        Место ребёнок = null;
        if (имя_ребёнка in дети)
        {
            ребёнок = дети[имя_ребёнка];
            дети.remove(имя_ребёнка);
            ребёнок.задать_родителя(null);
        }
        return ребёнок;
    }
    
    /++
    Функция, очищающая список потомков этого $(B Места).
    +/
    public void удалить_всех_детей(){
        foreach (имя_ребёнка, ребёнок; дети) 
            ребёнок.задать_родителя(null);
        дети = ПУСТОЙ_МАССИВ_ДЕТЕЙ.dup;    
    }
    
    protected Матрица4 _получить_полное_преобразование() {
        if (кэшированное_преобразование_устарело)
        {
            this.кэшированное_преобразование.матрица_преобразований(
                    позиция,
                    масштаб,
                    ориентация);
            кэшированное_преобразование_устарело = false;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("изменилось полное преобразование %s %s", 
                                                                 this.toString(), this.кэшированное_преобразование.toString()));
        }
        return кэшированное_преобразование;
    }
    
    /++
    Функция, задающая видимость этого $(B Места).
    
    Params:
        включено0 = результирующая видимость этого $(B Места)
    +/
    public void задать_видимость(bool включено0){
        включено = включено0;
    }

    /++
    Функция, возвращающая видимость этого $(B Места).
    +/
    public @property bool получить_видимость(){
        return включено;
    }

    /++
    Функция, переключающая видимость этого $(B Места).
    +/
    public void поменять_видимость(){
        включено = !включено;
    }
    
    /++
    Функция, добавляющая 3D-объект в это $(B Место) (в контейнер).

    Если переданный объект уже находится в каком либо $(B Месте), то функция не выполняет никаких действий.
    
    Params:
        до = добавляемый объект
    +/
    public void присоединить_объект(ДвижимыйОбъект до){
        if (до.родитель !is null) return;
        объекты[до.получить_имя] = до;
        до.задать_родителя(this);
    }
    
    /++
    Функция, возвращающая количество объектов контейнере.
    +/
    public @property short количество_объектов(){
        return to!(short) (объекты.length);
    }
    
    /++
    Функция, возвращающая объект из контейнера по его имени.
    
    Params:
        имя_объекта = имя искомого объекта
    +/
    public ДвижимыйОбъект получить_объект(string имя_объекта){
        return объекты[имя_объекта];
    }
    
    /++
    Функция, изымающая объект из контейнера.
    
    Params:
        до = изымаемый объект

    Returns:
        Изъятый объект.
    +/
    public ДвижимыйОбъект отсоединить_объект(ДвижимыйОбъект до){
        return отсоединить_объект(до.получить_имя);
    }
    
    /++
    Функция, изымающая объект по его имени из контейнера.
    
    Params:
        имя_объекта = имя изымаемого объекта

    Returns:
        Изъятый объект.
    +/
    public ДвижимыйОбъект отсоединить_объект(string имя_объекта){
        if (имя_объекта in объекты) {
            ДвижимыйОбъект результат = объекты[имя_объекта];
            объекты.remove(имя_объекта);
            результат.задать_родителя(null);
            return результат;
        }
        else
            return null;
    }
    
    /++
    Функция, очищающая контейнер от всех присоединённых объектов.
    +/
    public void отсоединить_все_объекты(){
        foreach (string имя_объекта, ДвижимыйОбъект до; объекты) 
            до.задать_родителя(null);
        объекты = ПУСТОЙ_МАССИВ_ОБЪЕКТОВ.dup;    
    }
    
    /++
    Функция рисования всех объектов, присоединённых к этому $(B Месту), и всех объектов, 
    присоединённых к потомкам этого $(B Места).

    Если видимость этого $(B Места) отключена, то функция не выполняет никаких действий.
    +/
    public void рисовать(){  
        if (включено){
            gl_сохранить_матрицу(); //glPushMatrix();
            
            gl_перемножить_матрицы(&(_получить_полное_преобразование().m[0])); // glMultMatrixf(&(_getFullTransform().m[0]));
            //gl.glTranslatef(mPosition.x, mPosition.y, mPosition.z);
            //gl.glScalef(mScale.x, mScale.y, mScale.z);
            //float rot[]=mOrientation.ToAngleAxis();
            //System.out.println("rot="+rot[0]+" "+rot[1]+" "+rot[2]+" "+rot[3]);
            //gl.glRotatef (rot[0]*RadtoGrad, rot[1], rot[2], rot[3]);
            
            //System.out.println("draw Node "+mName);
            
            //debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("рисование %s", this.toString()));
            
            foreach (объект; объекты.byValue()) 
                объект.рисовать();
            foreach (ребёнок; дети.byValue()) 
                ребёнок.рисовать();
            gl_восстановить_матрицу(); // glPopMatrix();
        }
    }
    
    /++
    Статистическая функция, возвращающая количество граней всех объектов, 
    присоединённых к этому $(B Месту), и всех объектов, 
    присоединённых к потомкам этого $(B Места).

    Если видимость этого $(B Места) отключена, то функция возвращает 0.
    +/
    public @property int количество_граней(){
        if (!включено) return 0;
        int результат = 0;
        //Set<String> keys = mObjects.keySet();
        foreach (до; объекты.byValue()) 
            результат += до.количество_граней;
        //keys = mChildren.keySet();
        foreach (ребёнок; дети.byValue()) 
            результат += ребёнок.количество_граней;
        return результат;
    }
    
    public Матрица4 получить_унаследованную_матрицу(){
        if (родитель is null)
            return _получить_полное_преобразование();
        else
            //return _getFullTransform().mul(mParent.getDerivedMatrix());
            return родитель.получить_унаследованную_матрицу() * _получить_полное_преобразование();
    }
    
    /// Вывод этого $(B Места) в строковое представление
    const public override string toString() {
        string строка_включено = включено ? "Включено" : "Выключено";
        string имя_родителя = родитель ? родитель.имя : "null";
        return "Место(" ~ имя ~ ": "  ~ строка_включено ~ " родитель=" ~ имя_родителя ~ " позиция=" ~ позиция.toString() ~ 
            " ориентация=" ~ ориентация.toString() ~ " масштаб=" ~ масштаб.toString() ~ ")";
    }

}
