/++
Модуль, представляющий классы работы с загружаемыми внешними ресурсами:
$(UL  
    $(LI  $(LINK2 #ИменаРесурсов, ИменаРесурсов) - контейнер всех имён ресурсов, доступных для загрузки  ) 
    $(LI  $(LINK2 #ЗагруженныеРесурсы, ЗагруженныеРесурсы) - абстрактный класс-родитель классов, 
        представляющих контейнеры с уже загруженными ресурсами конктретнгго типа )  
)

Также объявляется класс $(LINK2 #РесурсыИсключение, РесурсыИсключение) для исключений, 
которые могут возникнуть в результате загрузки ресурсов или их использования.
+/
module ddd.resources;

import std.file, std.string, std.path;
import my_utils.journal;

private enum имя_модуля = "ресурсы";

/++
Константа, в которой ключами являются все возможные типы ресурсов, 
а значениями - списки типов файлов, которые им соответствуют:
$(UL  
$(LI  меш: [".me"]) 
$(LI  скелет: [".sk"])
$(LI  картинка: [".jpg", ".png", ".tga", ".tif"]) 
$(LI  инфопанель: [".ip"])
$(LI  TTF_шрифт: [".ttf", ".fon"])
)
+/
enum auto ТИПЫ_РЕСУРСОВ=["меш":[".me"], 
                         "скелет":[".sk"],
                         "картинка":[".jpg", ".png", ".tga", ".tif"], // ".JPG", "PNG", ".TGA", ".TIF"], 
                         "объект":[".ob"],
                         "инфопанель":[".ip"],
                         "TTF_шрифт":[".ttf", ".fon"],
                         "texfontattr":[".fnattr"]
                         ];

/// Класс исключений, которые могут возникнуть при работе с внешними ресурсами
class РесурсыИсключение : Exception {
    this(string s) { 
        super(s); 
    }
};

/++
Класс - контейнер для всех имён ресурсов, которые доступны для загрузки из программы

Сканирует каждый каталог из переданного списка имён каталогов, и сохраняет все имена файлов,
которые там находятся и расширения которых соответствуют значениям константы $(LINK2 #ТИПЫ_РЕСУРСОВ, ТИПЫ_РЕСУРСОВ)
+/
class ИменаРесурсов {
    protected string[string][string] имена_ресурсов;
    protected string[] каталоги_ресурсов_;
    protected string[][string] типы_ресурсов;
    //protected string[string] loadedres;
    
    /++
    Конструктор, в который необходимо передать массив имён каталогов с ресурсами программы.

    Внутри вызывается метод $(LINK2 #загрузить_имена_файлов, загрузить_имена_файлов)
    +/
    this(string[] каталоги_ресурсов) {
        this.каталоги_ресурсов_=каталоги_ресурсов;
        типы_ресурсов=ТИПЫ_РЕСУРСОВ.dup;
        загрузить_имена_файлов(каталоги_ресурсов);
        в_журнал(имя_модуля, УровниЖурнала.Информация, 
            журнал_многострочное_форматирование(format("Загруженные имена ресурсов: %(\n%s %)" , имена_ресурсов))); 
    }

    /++
    Функция, сканирующая каталоги с ресурсами, и заполняющая контейнер именами файлов найденных ресурсов.
    +/
    void загрузить_имена_файлов(string[] каталоги_ресурсов) {
		string cwd=getcwd();
        int[string][string] типы;
        foreach (тип, расширения; типы_ресурсов) {
            foreach (расширение; расширения)
                типы[тип][расширение]=0;
            }
        foreach (имя_каталога; каталоги_ресурсов) 
            foreach (string имя_файла; dirEntries(имя_каталога, SpanMode.shallow)) {
                auto расширение=extension(имя_файла); // toLower(extension(имя_файла));
                auto основа_имени=baseName(имя_файла, расширение);
                if (расширение.length==0) continue;
                foreach (тип, расширения; типы_ресурсов) 
                    if (toLower(расширение) in типы[тип]) {
                        //имена_ресурсов[тип][основа_имени]= cwd ~ dirSeparator ~ имя_файла;
                        имена_ресурсов[тип][основа_имени] = имя_файла;
                        break;
                    } 
                    
                }
    }

    /++
    Возвращает имя файла ресурса (вместе с абсолютным путём к нему).

    Params:
        имя_ресурса = название ресурса, т.е. основа имени файла без расширения
        тип = название типа ресурса, должен быть одним из ключей константы $(LINK2 #ТИПЫ_РЕСУРСОВ, ТИПЫ_РЕСУРСОВ)

    Throws:
    $(UL  
    $(LI  $(B RangeError), если переданный тип отсутствует в константе $(LINK2 #ТИПЫ_РЕСУРСОВ, ТИПЫ_РЕСУРСОВ). )
    $(LI  $(LINK2 #РесурсыИсключение, РесурсыИсключение), если имя ресурса отсутствует среди загруженных имён. )
    )
    +/
    string получить_имя_файла_ресурса(string имя_ресурса, string тип){
        if (имя_ресурса in имена_ресурсов[тип])
            return имена_ресурсов[тип][имя_ресурса];
        else
            throw new РесурсыИсключение(тип ~ "[" ~ имя_ресурса ~ "]");
    }
}

/++
Абстрактный шаблон класса - контейнера загруженных ресурсов.

Потомки этого класса хранятся в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер)
в отдельной переменной для каждого типа ресурса.
Требуется создавать класс-потомок для каждого типа загружаемого ресурса.
Переопределять нужно только конструктор и функцию $(LINK2 #ЗагруженныеРесурсы.загрузить_один, загрузить_один).
+/
class ЗагруженныеРесурсы(T) { 

	// Ссылка на ранее загруженный массив с именами файлов
    protected ИменаРесурсов имена_ресурсов;
	// Сам контейнер
	protected T[string] ресурсы; 

    /++
    Конструктор, принимающий ссылку на объект типа $(LINK2 #ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов.
    +/
    this(ИменаРесурсов имена_ресурсов0) {
        this.имена_ресурсов=имена_ресурсов0;
    };

    /++
    Функция, загружающая в контейнер несколько ресурсов. 

    Params:
        загружаемые_имена_ресурсов = массив имён ресурсов, которые требуется загрузить
        первично = Если этот параметр равен $(B true), и требуемый ресурс уже загружен, то повторная загрузка не производится. 
        В случае значения $(B false) ресурс будет загружен вне зависимости от того, был ли он загружен ранее.
    +/
    void загрузить(string[] загружаемые_имена_ресурсов, bool первично=true) {
        foreach (string имя; загружаемые_имена_ресурсов) {
            if (первично && (имя in ресурсы))
                continue;
            загрузить_один(имя);
        }
    }

    /++
    Абстрактная функция загрузки одного файла с ресурсом. Требуется переопределить в каждом из потомков.
    +/
	abstract protected bool загрузить_один(string имя);

    /++
    Функция, возвращающая ссылку на ресурс по его имени. Если ресурс в контейнере отсутствует, 
    делается попытка загрузить его из файла. 

    Returns:
        Ссылка на ресурс, если он присутствовал в контейнере или успешно загрузился. $(B null), если загрузить файл ресурса не удалось.
    +/
    T получить(string имя) {
        if (имя in ресурсы)
            return ресурсы[имя];
        if (загрузить_один(имя))
            return ресурсы[имя];
        в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Отсутствует ресурс с именем: %s" , имя)); 
        return null;
    }

    /++
    Функция, загружающая в контейнер несколько ресурсов вне зависимости от того, были ли они загружены ранее. 


    Params:
        загружаемые_имена_ресурсов = массив имён ресурсов, которые требуется загрузить
    +/
    void перезагрузить(string[] загружаемые_имена_ресурсов) {
        загрузить(загружаемые_имена_ресурсов, false);
    }
	
}
