/++
Модуль, определяющий класс $(LINK2 #Скелет, Скелет), который позволяет выполнять скелетную анимацию
на 3D-объектах. Пользователю обычно требуется только создать скелет (посредством загрузки из файла) и
присоединить его к мешу. Все дальнейшие действия, связанные с анимацими, выполняются через 
анимированный меш.

В модуле определено несколько вспомогательных структур, используемых внутри класса $(LINK2 #Скелет, Скелет).
Также определён класс ресурсов $(LINK2 #ЗагруженныеСкелеты, ЗагруженныеСкелеты), позволяющий получать объекты 
$(LINK2 #Скелет, Скелет) из файлов-ресурсов специального бинарного формата с расширением ".sk". 
Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты класса $(LINK2 #Скелет, Скелет) через него.
+/
module ddd.skelet;

import std.stdio, std.array, std.typecons, std.string, std.file, std.math, std.conv;

import ddd.vector, ddd.quaternion, ddd.resources;
import my_utils.journal;

enum имя_модуля = "скелет";
enum ТИП_РЕСУРСА = "скелет";

enum РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ = 6;
enum РАЗМЕР_SHORT = 2;
enum РАЗМЕР_FLOAT = 4;


enum byte[] ЗАГОЛОВОК_СКЕЛЕТА_2_ВЕРСИИ = [68, 68, 68, 32, 2, 0]; //, 0, 0];

/++
Константы, определяющие 10 типов кривых анимации. Их можно объединять через операцию |.
+/
enum ТипыКривыхАнимации : ushort    {
    ничего = 0,
    положение_x =  1,
    положение_y =  2,
    положение_z =  4,
    ориентация_w = 8,
    ориентация_x = 16,
    ориентация_y = 32,
    ориентация_z = 64,
    масштаб_x =   128,
    масштаб_y =   256,
    масштаб_z =   512,
}

/++
Константы, определяющие типы изменений при анимации, т.е. положение, ориентация или масштаб.
Их можно объединять через операцию |.
+/
enum ТипыИзменения : ushort {
    ничего = 0,
    положение = 1,
    ориентация = 2,
    масштаб = 4
}    

/+
Вспомогательный тип кортежа, используемый в структуре $(LINK2 #Анимация, Анимация)
+/
alias ДиспетчерКривых = Tuple!(ushort, "номер_кости", ushort, "тип_кривой"); //, ushort, "номер_кривой");

/++
Класс исключений, которые могут случиться при работе объектов класса $(LINK2 #Скелет, Скелет).
+/
class СкелетИсключение : Exception {
    this(string s) { 
        super(s); 
    }
};

/++
Структура, определяющая начальное положение и ориентацию кости скелета.
+/
struct НачальнаяКость {
    Вектор3 положение;
    Кватернион ориентация;
}

/++
Структура, определяющая анимацию кости скелета.
+/
struct АнимированнаяКость {
    //ushort наличие_кривых;
    float[ushort] кривые;
}

/++
Структура, в которой содержится информация о единице анимации (или, иначе, о "действии").
В анимации определены номера начального и конечного кадра, задействуется определённый набор костей 
(какие именно, задаётся в массиве $(B диспетчер_кривых)), и массив кривых анимации. Размеры массивов
должны совпадать, одинаковые индексы массивов относятся к анимации одной и той же кости.
+/
struct Анимация {
    /// Номер начального кадра
    short стартовый_кадр;
    /// Номер последнего кадра
    short конечный_кадр;
    /++
    Массив, в котором определяется к какой кости (по номеру) в скелете относится каждая из кривых,
    и какого типа эта кривая (в соответствии с перечислением $(LINK2 #ТипыКривыхАнимации, ТипыКривыхАнимации)).
    +/
    ДиспетчерКривых[] диспетчер_кривых;
    /// Массив с кривыми анимации
    float[][] кривые;

    /// Вывод структуры в строковое представление. Осторожно, может быть очень длинный результат!
    string toString() {
        string строка_диспетчер = "диспетчер=[";
        foreach(номер_тип; диспетчер_кривых) 
            строка_диспетчер ~= format("(%s, %s),", номер_тип.номер_кости, номер_тип.тип_кривой);
        return "Анимация(" ~ format("(%s,  %s), ", стартовый_кадр, конечный_кадр) ~
            строка_диспетчер[0..$-1] ~ format("], кривые=[%([%(%.2g, %)],%))])", кривые);
    }
}

/++
Структура, определяющая текущее положение отдельной кости
+/
struct СостояниеАнимацииКости {
    /// Положение кости без применения каких-либо анимаций
    НачальнаяКость начальное_состояние;
    /// Какому типу изменений подверглась кость (см. константу $(LINK2 #ТипыИзменения, ТипыИзменения))
    ushort типы_изменения;
    /// Вектор сдвига
    Вектор3 изменение_положения;
    /// Кватернион поворота
    Кватернион изменение_ориентации;
    /// Вектор масштабирования
    Вектор3 изменение_масштаба;

    /// Вывод структуры в строковое представление.
    string toString() {
        return "СостояниеАнимацииКости((" ~ начальное_состояние.положение.toString() ~ ", " ~
            начальное_состояние.ориентация.toString() ~ "), " ~ to!string(типы_изменения) ~ ", " ~ изменение_положения.toString() ~
            ", " ~ изменение_ориентации.toString() ~ ", " ~ изменение_масштаба.toString() ~ ")";
    }
}    

/++
Константа типа $(LINK2 #СостояниеАнимацииКости, СостояниеАнимацииКости), определяющая вуртуальную кость
с нулевым воздействием на меш, т.е. кость, стоящую в точке с нулевыми координатами, без поворота,
и не подвергнувшуюся никаким анимациям.
+/
const СостояниеАнимацииКости СОСТОЯНИЕ_БЕЗ_ИЗМЕНЕНИЙ = 
                       СостояниеАнимацииКости(НачальнаяКость(Вектор3.НОЛЬ, Кватернион.НОЛЬ),
                                              ТипыИзменения.ничего,
                                              Вектор3(Вектор3.НОЛЬ),
                                              Кватернион(Кватернион.НОЛЬ),
                                              Вектор3(Вектор3.ЕДИНИЦА_МАСШТАБА));

/// Вывод ассоциативного массива с состояниями анимации костей в строковое представление
string состояние_анимации_в_строку(СостояниеАнимацииКости[string] состояние_анимации) {
    string результат;
    foreach(имя_кости, состояние_кости; состояние_анимации)
        результат ~= имя_кости ~ ": " ~ состояние_кости.toString() ~ "\n";
    return результат;
}

/++
Класс, определяющий упорядоченный набор костей и их анимаций. Используется для анимаций 3D-объектов,
в данный момент такие анимации возможно применять только к классу 
$(LINK2 mesh.html#МешСоСкелетом, МешСоСкелетом).
+/
class Скелет {
    protected string[] имена_костей;
    protected НачальнаяКость[] кости;
    protected Анимация[string] анимации;
    protected short[] обратная_иерархия;
    protected ushort[][ushort] прямая_иерархия;
    protected ushort[] корни_иерархии;

    /// Коструктор, принимающий имя файла с данными скелета и его анимаций.
    this(string имя_файла_скелета) {
        в_журнал(имя_модуля, УровниЖурнала.Информация, format("Начинаем загрузку скелета %s", имя_файла_скелета));
        enum РАЗМЕР_ЗАГОЛОВКА = 10;
        enum РАЗМЕР_НАЧАЛЬНОЙ_КОСТИ = 40;
        enum РАЗМЕР_ЗАГОЛОВКА_АНИМАЦИИ = 4;
        auto байты1 = read(имя_файла_скелета);
        if (байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ] != ЗАГОЛОВОК_СКЕЛЕТА_2_ВЕРСИИ) {
            writeln(байты1[0..РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ]);
            writeln("Случилась ошибка заголовка при чтении файла скелета");
            throw new СкелетИсключение("Заголовок файла не соответствует требуемому для скелета!"); 
        }
        ushort[] длины = cast(ushort[]) байты1[РАЗМЕР_ЗАГОЛОВКА_ВЕРСИИ..РАЗМЕР_ЗАГОЛОВКА];
        ushort количество_костей = длины[0];
        ushort количество_анимаций = длины[1];

        size_t место_чтения = РАЗМЕР_ЗАГОЛОВКА;
        ushort длина_имени;
        for (int i=0; i<количество_костей; i++) {
            длина_имени = (cast(ushort[]) байты1[место_чтения .. (место_чтения+РАЗМЕР_SHORT)])[0];
            место_чтения += РАЗМЕР_SHORT;
            имена_костей ~= cast(string) байты1[место_чтения .. (место_чтения+длина_имени)];
            место_чтения += длина_имени;
        }

        обратная_иерархия = cast(short[]) байты1[место_чтения .. место_чтения+РАЗМЕР_SHORT*количество_костей];
        место_чтения += РАЗМЕР_SHORT*количество_костей;
        for (int i=0; i<количество_костей; i++) {
            float[] данные = cast(float[]) байты1[место_чтения .. место_чтения+РАЗМЕР_НАЧАЛЬНОЙ_КОСТИ];
            место_чтения+=РАЗМЕР_НАЧАЛЬНОЙ_КОСТИ;
            кости ~= НачальнаяКость(Вектор3(данные[0..3]), Кватернион(данные[6..10]));
        }

        string[] имена_анимаций;
        for (int i=0; i<количество_анимаций; i++) {
            длина_имени = (cast(ushort[]) байты1[место_чтения .. (место_чтения+РАЗМЕР_SHORT)])[0];
            место_чтения += РАЗМЕР_SHORT;
            имена_анимаций ~= cast(string) байты1[место_чтения .. (место_чтения+длина_имени)];
            место_чтения += длина_имени;
        }
        //writeln("имена_анимаций=", имена_анимаций);
        //writeln("место_чтения после имена_анимаций =", место_чтения);
        
        for (int i=0; i<количество_анимаций; i++) {
            длины = cast(ushort[]) байты1[место_чтения .. место_чтения+РАЗМЕР_ЗАГОЛОВКА_АНИМАЦИИ];
            ushort длина_диспетчера = длины[0];
            ushort количество_кадров = длины[1];
            место_чтения += РАЗМЕР_ЗАГОЛОВКА_АНИМАЦИИ;
            short[] пределы_кадров = cast(short[]) байты1[место_чтения .. место_чтения+РАЗМЕР_SHORT*2];
            short стартовый_кадр = пределы_кадров[0];
            short конечный_кадр = пределы_кадров[1];
            место_чтения += РАЗМЕР_SHORT*2;
            size_t длина_буфера = РАЗМЕР_SHORT*2*длина_диспетчера;
            ushort[] массив_диспетчера = cast(ushort[]) байты1[место_чтения .. место_чтения+длина_буфера];
            место_чтения += длина_буфера;
            ДиспетчерКривых[] диспетчер;
            for (int j=0; j<длина_диспетчера; j++)
                диспетчер ~= ДиспетчерКривых(массив_диспетчера[j*2], массив_диспетчера[j*2+1]);
            float[][] кривые = uninitializedArray!(float[][])(количество_кадров, длина_диспетчера);
            длина_буфера = РАЗМЕР_FLOAT*количество_кадров*длина_диспетчера;
            float[] буфер_кривых = cast(float[]) байты1[место_чтения .. место_чтения+длина_буфера];
            //writeln(имена_анимаций[i], ":\n", буфер_кривых);
            место_чтения += длина_буфера;
            for (int k=0; k<количество_кадров; k++) 
                for (int j=0; j<длина_диспетчера; j++)
                    кривые[k][j] = буфер_кривых[k*длина_диспетчера+j];
            анимации[имена_анимаций[i]] = Анимация(стартовый_кадр, конечный_кадр, диспетчер, кривые);
        }
        _построение_прямой_иерархии();
    }

    private void _построение_прямой_иерархии() {
        for (ushort номер_кости=0; номер_кости< обратная_иерархия.length; номер_кости++) {
            //if (!(номер_кости in прямая_иерархия))
            //    прямая_иерархия[номер_кости] = [];
            if (обратная_иерархия[номер_кости] < 0)
                корни_иерархии ~= номер_кости;
            else {
                //ushort номер_родителя = обратная_иерархия[номер_кости];
                прямая_иерархия[обратная_иерархия[номер_кости]] ~= номер_кости;
            }
                
        }
    }

    /// Вывод информации о скелете в строковое представление
    override string toString() {
        return "Скелет(\n" ~ format("костей=%s,  анимаций=%s,\n", кости.length, анимации.length) ~
            format("имена костей=%s\n", имена_костей) ~ format("кости=%s\n", кости) ~
            format("иерархия=%s, %s\n", корни_иерархии, прямая_иерархия) ~ format("анимации=%s\n", анимации);
    }

    /++
    Функция, возвращающая состояние костей в определённый момент анимации

    Params:
        название_анимации = Имя используемой анимации (или, иначе, "действия").
        положение_кадра = Для какого кадра требуется получить состояние. Может быть нецелым числом, 
        в этом случае используется линейная интерполяция.

    Throws:
        $(B RangeError), если анимация с таким именем отсутсвует.
    +/
    СостояниеАнимацииКости[string] получить_состояние_анимации(string название_анимации, float положение_кадра) {
        auto состояние = _получить_состояние_анимации_float(название_анимации, положение_кадра);
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("состояние скелета до пересчёта = %s", состояние));
        return _пересчитать_состояние_анимации(состояние);
    }

    private СостояниеАнимацииКости[ushort] _получить_состояние_анимации_float(string название_анимации, 
                                                                              float положение_кадра) {
        Анимация анимация = анимации[название_анимации]; // Выбросит исключение, если такой анимации нет.
        //writeln(название_анимации, ": (", анимация.стартовый_кадр, ", ", анимация.конечный_кадр,
        //    ")[", положение_кадра, "]");
        if (положение_кадра <= анимация.стартовый_кадр) 
            return _получить_состояние_анимации(анимация, анимация.стартовый_кадр);
        if (положение_кадра >= анимация.конечный_кадр) 
            return _получить_состояние_анимации(анимация, анимация.конечный_кадр);
        short кадр1 = to!short(floor(положение_кадра));
        short кадр2 = to!short(ceil(положение_кадра));
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("положение_кадра = %s, " ~ 
            "кадр1 = %s, кадр2 = %s", положение_кадра, кадр1, кадр2));
        if (кадр1 == кадр2)
            return _получить_состояние_анимации(анимация, кадр1);
        auto состояние1 = _получить_состояние_анимации(анимация, кадр1);
        auto состояние2 = _получить_состояние_анимации(анимация, кадр2);
        return интерполяция_анимации(состояние1, состояние2, положение_кадра-кадр1);
    }

    private СостояниеАнимацииКости[ushort] интерполяция_анимации(СостояниеАнимацииКости[ushort] состояние1, 
                                                         СостояниеАнимацииКости[ushort] состояние2, 
                                                         float междукадрие) {
        float t1 = (1.0 - междукадрие);
        float t2 = междукадрие;
        СостояниеАнимацииКости[ushort] результат = состояние1;
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("результат до интреполяции = %s", результат));
        foreach (номер_кости, состояние_кости; состояние1) {
            if (состояние_кости.типы_изменения & ТипыИзменения.положение)
                результат[номер_кости].изменение_положения = (состояние_кости.изменение_положения*t1) 
                    + (состояние2[номер_кости].изменение_положения*t2);
            if (состояние_кости.типы_изменения & ТипыИзменения.ориентация)
                результат[номер_кости].изменение_ориентации = 
                    состояние_кости.изменение_ориентации.линейная_интерполяция(состояние2[номер_кости].изменение_ориентации, междукадрие);
            if (состояние_кости.типы_изменения & ТипыИзменения.масштаб)
                результат[номер_кости].изменение_масштаба = (состояние_кости.изменение_масштаба*t1) 
                    + (состояние2[номер_кости].изменение_масштаба*t2);
        }
        debug в_журнал(имя_модуля, УровниЖурнала.Отладка, format("результат после интреполяции = %s", результат));
        return результат;
    }
    
    private СостояниеАнимацииКости[ushort] _получить_состояние_анимации(Анимация анимация, short кадр) {
        СостояниеАнимацииКости[ushort] результат;
        int индекс_кадра = кадр - анимация.стартовый_кадр;
        foreach (индекс, тип_изменения; анимация.диспетчер_кривых) {
            //writeln(индекс, ", (", тип_изменения.номер_кости, ", ", тип_изменения.тип_кривой, "), ", анимация.кривые[индекс_кадра][индекс]);
            ushort номер_кости = тип_изменения.номер_кости;
            if (!(номер_кости in результат)) {
                результат[номер_кости] = СостояниеАнимацииКости(кости[номер_кости],
                                                              ТипыИзменения.ничего,
                                                              Вектор3(Вектор3.НОЛЬ),
                                                              Кватернион(Кватернион.НОЛЬ),
                                                              Вектор3(Вектор3.ЕДИНИЦА_МАСШТАБА));
            }
            switch(тип_изменения.тип_кривой) {
                case ТипыКривыхАнимации.положение_x:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.положение;
                    результат[номер_кости].изменение_положения.x = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.положение_y:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.положение;
                    результат[номер_кости].изменение_положения.y = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.положение_z:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.положение;
                    результат[номер_кости].изменение_положения.z = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.ориентация_w:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.ориентация;
                    результат[номер_кости].изменение_ориентации.w = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.ориентация_x:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.ориентация;
                    результат[номер_кости].изменение_ориентации.x = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.ориентация_y:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.ориентация;
                    результат[номер_кости].изменение_ориентации.y = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.ориентация_z:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.ориентация;
                    результат[номер_кости].изменение_ориентации.z = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.масштаб_x:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.масштаб;
                    результат[номер_кости].изменение_масштаба.x = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.масштаб_y:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.масштаб;
                    результат[номер_кости].изменение_масштаба.y = анимация.кривые[индекс_кадра][индекс];
                    break;
                case ТипыКривыхАнимации.масштаб_z:
                    результат[номер_кости].типы_изменения |= ТипыИзменения.масштаб;
                    результат[номер_кости].изменение_масштаба.z = анимация.кривые[индекс_кадра][индекс];
                    break;
                default : 
                    break;
            }
                
        }
        return результат;
    }

    
    private СостояниеАнимацииКости[string] _пересчитать_состояние_анимации(СостояниеАнимацииКости[ushort] состояние) {
        СостояниеАнимацииКости[string] результат;
        for (ushort номер_кости=0; номер_кости<кости.length; номер_кости++)
            if (номер_кости in состояние)
                результат[имена_костей[номер_кости]] = состояние[номер_кости];
            else 
                результат[имена_костей[номер_кости]] = СостояниеАнимацииКости(кости[номер_кости],
                                                              ТипыИзменения.ничего,
                                                              Вектор3(Вектор3.НОЛЬ),
                                                              Кватернион(Кватернион.НОЛЬ),
                                                              Вектор3(Вектор3.ЕДИНИЦА_МАСШТАБА));
        return результат;
    }
    
    private СостояниеАнимацииКости[string] _пересчитать_состояние_анимации1(СостояниеАнимацииКости[ushort] состояние) {
        СостояниеАнимацииКости[string] результат;
        СостояниеАнимацииКости текущее_состояние;
        foreach (номер_кости; корни_иерархии) {
            if (номер_кости in состояние) {
                текущее_состояние = состояние[номер_кости];
                текущее_состояние.изменение_ориентации = кости[номер_кости].ориентация * 
                                          состояние[номер_кости].изменение_ориентации *
                                          кости[номер_кости].ориентация.обратить();
            }
            else
                текущее_состояние = СостояниеАнимацииКости(кости[номер_кости],
                                                              ТипыИзменения.ничего,
                                                              Вектор3(Вектор3.НОЛЬ),
                                                              Кватернион(Кватернион.НОЛЬ),
                                                              Вектор3(Вектор3.ЕДИНИЦА_МАСШТАБА));
            результат[имена_костей[номер_кости]] = текущее_состояние;
            foreach (номер_дочерней_кости; прямая_иерархия[номер_кости])
                _рекурсивный_пересчёт_состояния(номер_дочерней_кости, состояние, 
                                                текущее_состояние, результат);
        }
        return результат;
    }

    private void _рекурсивный_пересчёт_состояния(ushort номер_кости, 
                                                 СостояниеАнимацииКости[ushort] состояние, 
                                                 СостояниеАнимацииКости состояние_родителя, 
                                                 ref СостояниеАнимацииКости[string] результат) {
        //СостояниеАнимацииКости текущее_состояние;
        Вектор3 новое_положение = кости[номер_кости].положение + состояние_родителя.изменение_положения;
        Вектор3 вектор_от_родителя = кости[номер_кости].положение - кости[обратная_иерархия[номер_кости]].положение;
        вектор_от_родителя *= состояние_родителя.изменение_масштаба;
        новое_положение += состояние_родителя.изменение_ориентации * вектор_от_родителя - вектор_от_родителя;
        //writeln("кость ", номер_кости, ": новое_положение=", новое_положение);
        Вектор3 новое_изменение_положения = новое_положение - кости[номер_кости].положение;
        Кватернион новое_изменение_ориентации = Кватернион(состояние_родителя.изменение_ориентации);
        Вектор3 новое_изменение_масштаба = Вектор3(состояние_родителя.изменение_масштаба);
        ushort новые_типы_изменения = состояние_родителя.типы_изменения;
        if (номер_кости in состояние) {
            новые_типы_изменения = новые_типы_изменения || состояние[номер_кости].типы_изменения;
            новое_изменение_положения += состояние[номер_кости].изменение_положения;
            новое_изменение_ориентации *= кости[номер_кости].ориентация * 
                                          состояние[номер_кости].изменение_ориентации *
                                          кости[номер_кости].ориентация.обратить();
            новое_изменение_масштаба *= состояние[номер_кости].изменение_масштаба;
        }
        if (!новое_изменение_положения.нулевая_длина())
            новые_типы_изменения = новые_типы_изменения || ТипыИзменения.положение;
        СостояниеАнимацииКости текущее_состояние = СостояниеАнимацииКости(кости[номер_кости],
                                                          новые_типы_изменения,
                                                          новое_изменение_положения,
                                                          новое_изменение_ориентации,
                                                          новое_изменение_масштаба);
        результат[имена_костей[номер_кости]] = текущее_состояние;
        if (номер_кости in прямая_иерархия)
            foreach (номер_дочерней_кости; прямая_иерархия[номер_кости])
                _рекурсивный_пересчёт_состояния(номер_дочерней_кости, состояние, 
                                                текущее_состояние, результат);
    }

    /++
    Вывод информации о состояниях костей при применении двух анимаций одновременно.

    Params:
        название_анимации1 = Имя первой используемой анимации (или, иначе, "действия").
        положение_кадра1 = Для какого кадра первой анимации требуется получить состояние. 
        название_анимации2 = Имя второй используемой анимации (или, иначе, "действия").
        положение_кадра2 = Для какого кадра второй анимации требуется получить состояние. 
        коэффициент = Если вычисляется состояние кости, которая участвует в обеих анимациях,
        то этот параметр используется как коэффициент интерполяции между ними. Если кость,
        для которой вычисляется состояние, используется только в одной из анимаций, то 
        этот параметр не используется.


    Throws:
        $(B RangeError), если название_анимации1 или название_анимации2 отсутствует в списке 
        анимаций скелета.
    
    +/
    СостояниеАнимацииКости[string] объединить_состояния_анимации(string название_анимации1, 
                                                                 float положение_кадра1,
                                                                 string название_анимации2, 
                                                                 float положение_кадра2,
                                                                 float коэффициент) {
        СостояниеАнимацииКости[string] результат;
        auto анимация1 = получить_состояние_анимации(название_анимации1, положение_кадра1);
        auto анимация2 = получить_состояние_анимации(название_анимации2, положение_кадра2);
        float t1 = 1 - коэффициент;
        float t2 = коэффициент;
        foreach (имя_кости; имена_костей) {
            if (!((имя_кости in анимация1) is null) & !((имя_кости in анимация2) is null)) {
                результат[имя_кости] = анимация1[имя_кости];
                результат[имя_кости].типы_изменения |= анимация1[имя_кости].типы_изменения;
                результат[имя_кости].изменение_положения = (анимация1[имя_кости].изменение_положения * t1 + 
                                                            анимация2[имя_кости].изменение_положения * t2);
                результат[имя_кости].изменение_масштаба = (анимация1[имя_кости].изменение_масштаба * t1 + 
                                                            анимация2[имя_кости].изменение_масштаба * t2);
                результат[имя_кости].изменение_ориентации = анимация1[имя_кости].изменение_ориентации.сферическая_интерполяция(
                    анимация2[имя_кости].изменение_ориентации, коэффициент);
            }
            else if (имя_кости in анимация1)
                результат[имя_кости] = анимация1[имя_кости];
            else if (имя_кости in анимация2)
                результат[имя_кости] = анимация2[имя_кости];
        }
        return результат;
    }


}

/++
Класс-контейнер ресурсов объектов класса $(LINK2 #Скелет, Скелет).

Экземпляр этого класса присутствует в объекте $(LINK2 ddd_manager.html#DDDМенеджер, DDDМенеджер).
Рекомендуется получать объекты класса $(LINK2 #Скелет, Скелет) через него.
+/
class ЗагруженныеСкелеты : ЗагруженныеРесурсы!(Скелет) {

    /++
    Конструктор, принимающий ссылку на объект типа $(LINK2 resources.html#ИменаРесурсов, ИменаРесурсов), 
    в котором уже должны храниться имена всех доступных файлов ресурсов.
    +/
    this(ИменаРесурсов имена_ресурсов0) {
        super(имена_ресурсов0);
    }

    /++
    Функция загрузки в контейнер одного файла с ресурсом. 

    Returns:
        true, если файл удачно загрузился, false в противном случае.
    +/
    override protected bool загрузить_один(string имя) {
        string имя_файла_скелета;
        Скелет новый_скелет;
        try 
            имя_файла_скелета = имена_ресурсов.получить_имя_файла_ресурса(имя, ТИП_РЕСУРСА);
        catch (РесурсыИсключение) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Отсутствует файл скелета %s", имя));
            return false;
        }
        try 
            новый_скелет = new Скелет(имя_файла_скелета);
        catch (Throwable) {
            в_журнал(имя_модуля, УровниЖурнала.Ошибка, format("Ошибка в файле скелета %s", имя_файла_скелета));
            return false;
        }
        ресурсы[имя] = новый_скелет;
        return true;
    }
    
}
