﻿/++
`journal` - модуль журналирования, поддерживающий различные уровни для отдельных вызывающих модулей.

Для создания журнала требуется определить общий уровень (используемый по-умолчанию). 
Для отдельных модулей можно определить собственный уровень в ассоциативном массиве.
Вместе с именем файла журнала эти параметры нужно передать в функцию инициализации, 
либо в $(LINK2 #инициализация_журнала, инициализация_журнала) (с параметрами типа $(LINK2 #УровниЖурнала, УровниЖурнала)), 
либо в $(LINK2 #строковая_инициализация_журнала, строковая_инициализация_журнала) (соответственно, со строковыми параметрами). После этого можно
писать в журнал с помощью функции $(LINK2 #в_журнал, в_журнал). 

В случае многострочных записей перед передачей их в функцию $(LINK2 #в_журнал, в_журнал) рекомендуется 
пропустить их через функцию $(LINK2 #журнал_многострочное_форматирование, журнал_многострочное_форматирование),
чтобы все строки аккуратно располагались друг под другом в графе "Сообщение".


+/
module my_utils.journal;

///
unittest {
    import core.thread, std.stdio;
    import std.algorithm : endsWith;
    import core.time : msecs;
    
    enum ФАЙЛ_ЖУРНАЛА = "журнал0.txt";
    УровниЖурнала[string] уровни_модулей1 = ["журнал": УровниЖурнала.Информация];

    // Открываем журнал и пишем в него
    инициализация_журнала(ФАЙЛ_ЖУРНАЛА, УровниЖурнала.Отладка, уровни_модулей1);
    // Эта запись не должна попасть в журнал
    в_журнал("журнал", УровниЖурнала.Отладка, "запись 1"); 
    в_журнал("журнал", УровниЖурнала.Предупреждение, "запись 2"); 
    Thread.sleep(msecs(100));
    в_журнал("не журнал", УровниЖурнала.Информация, "запись 3");
    остановить_журнал();

    // Теперь читаем записанное в журнал
    Thread.sleep(msecs(200));   // На сохранение в файл нужно время
    File файл_журнала = File(ФАЙЛ_ЖУРНАЛА, "r");
    for (int i=0; i<4; i++) файл_журнала.readln();  // Заголовок пропускаем
    assert(файл_журнала.readln().endsWith("запись 2\n"));
    assert(файл_журнала.readln().endsWith("запись 3\n"));
    assert(файл_журнала.readln() == "");
    assert(файл_журнала.eof());
    файл_журнала.close();
}    

import std.stdio, std.datetime, std.concurrency, std.string, std.conv, std.container, std.utf;
import std.range;
import core.exception;


/++ Перечисление с уровнями журналирования  +/
enum УровниЖурнала {КритическаяОшибка, Ошибка, Предупреждение, Информация, Отладка};

// Константы 
private int РАЗМЕР_БУФЕРА = 200;
private int РАЗМЕР_ВЫГРУЗКИ = 50;
private int РАЗМЕР_ПОЧТОВОГО_ЯЩИКА = 300;
private float МНОЖИТЕЛЬ_ВРЕМЕНИ = 10000000.0;
private enum ОБЩИЙ_УРОВЕНЬ_ПО_УМОЛЧАНИЮ = УровниЖурнала.Предупреждение; 
private enum ИМЯ_ПОТОКА_ЖУРНАЛИРОВАНИЯ = "поток_журналирования";

/++ То, как в журнале отображается уровень конкретного сообщения +/
enum string[УровниЖурнала] УровеньСообщения = [
    УровниЖурнала.КритическаяОшибка: "Крит.ошибка",
    УровниЖурнала.Ошибка           : "Ошибка     ",
    УровниЖурнала.Предупреждение   : "Предупрежд.",
    УровниЖурнала.Информация       : "Информация ",
    УровниЖурнала.Отладка          : "Отладка    "
    ];

	
/++
Основная функция инициализации системы журналирования

Создаёт поток, ведущий журнал

Params:
    файл_журнала   = имя файла журнала
    общий_уровень = текущий уровень журналирования для большинства модулей
    уровни_модулей0   = ассоциативный массив уровней для отдельных модулей 
+/
void инициализация_журнала(string файл_журнала, 
                           УровниЖурнала общий_уровень, 
						   УровниЖурнала[string] уровни_модулей0) {
    with (параметры_журнала) {
        текущий_общий_уровень=общий_уровень;
        //writeln("  текущий_общий_уровень=", текущий_общий_уровень);
        уровни_модулей=cast(shared(УровниЖурнала[string])) уровни_модулей0;
        //writeln("присвоили переменные");
        //объект_журнала=new ЖурналОбъект(logfilename, 40, 20);
        auto tid_поток_журналирования1=spawn(&функция_потока_журнала, файл_журнала, РАЗМЕР_БУФЕРА, РАЗМЕР_ВЫГРУЗКИ);
        //writeln("запущен поток");
        //bool результат_регистрации=register(ИМЯ_ПОТОКА_ЖУРНАЛИРОВАНИЯ, tid_поток_журналирования1);
        register(ИМЯ_ПОТОКА_ЖУРНАЛИРОВАНИЯ, tid_поток_журналирования1);
        tid_поток_журналирования=tid_поток_журналирования1;
        //writeln("присвоили переменные");
        setMaxMailboxSize(tid_поток_журналирования1, РАЗМЕР_ПОЧТОВОГО_ЯЩИКА, OnCrowding.block);
        //тиков_в_секунду=TickDuration.ticksPerSec;
        long тики=текущее_время();
        float секунды=тики/МНОЖИТЕЛЬ_ВРЕМЕНИ;
		
		tid_поток_журналирования1.send(thisTid, тики, format("%10.5f  ", секунды) ~ заголовок());
		
    }                    
}

/++
Массив названий модулей, у которых при конвертации в функции
$(LINK2 #конвертация_уровней_модулей, конвертация_уровней_модулей) произошла ошибка
+/
string[] ошибки_конвертации_уровней_модулей;

/++
Функция конвертации из массива со строковым написанием уровней
в массив с уровнями $(LINK2 #УровниЖурнала, УровниЖурнала), 
пригодный для функции инициализация_журнала

Params:
    уровни_модулей_строки = уровни модулей в строковом написании

Returns:
    уровни модулей с типом $(B УровниЖурнала)[string]

Note:
    побочное действие: в случае ошибок названия соответствующих модулей
    сохраняются в массиве $(LINK2 #ошибки_конвертации_уровней_модулей, ошибки_конвертации_уровней_модулей)
+/
auto конвертация_уровней_модулей(string[string] уровни_модулей_строки) {
    УровниЖурнала[string] результат;
    foreach (модуль, уровень; уровни_модулей_строки) {
        try 
            результат[модуль]=ПреобразовательУровнейЖурнала[уровень];
        catch (RangeError)
            ошибки_конвертации_уровней_модулей ~= модуль;
    }
    return результат;
}

/++
Функция инициализации системы журналирования со строковыми параметрами


Преобразует параметры со строковыми уровнями журналирования в значения с типом 
$(LINK2 #УровниЖурнала, УровниЖурнала), 
затем вызывается $(LINK2 #инициализация_журнала, инициализация_журнала). 

Если конвертация общего уровня журналирования прошла с ошибкой, используется уровень по-умолчанию.
Если были ошибки в строках для каких-либо модулей, то для них будет использоваться общий уровень.
Информация об этих ошибках записывается в журнал. 


Params:
    файл_журнала   = имя файла журнала
    общий_уровень_строка = текущий уровень журналирования для большинства модулей
    уровни_модулей_строки = ассоциативный массив уровней для отдельных модулей 
+/
void строковая_инициализация_журнала(string файл_журнала, 
									 string общий_уровень_строка, 
									 string[string] уровни_модулей_строки) {
    УровниЖурнала[string] уровни_модулей=конвертация_уровней_модулей(уровни_модулей_строки);
    УровниЖурнала общий_уровень=ОБЩИЙ_УРОВЕНЬ_ПО_УМОЛЧАНИЮ;
    bool ошибка_общего_уровня=false;
    
    if ((общий_уровень_строка in ПреобразовательУровнейЖурнала) !is null)
        общий_уровень=ПреобразовательУровнейЖурнала[общий_уровень_строка];
    else
        ошибка_общего_уровня=true;

    
    инициализация_журнала(файл_журнала, общий_уровень, уровни_модулей);
    if (ошибка_общего_уровня)
        tid_поток_журналирования.send(thisTid, текущее_время(), 
             "Конвертация общего уровня журналирования прошла с ошибкой, используется уровень по-умолчанию. \n");
    if (ошибки_конвертации_уровней_модулей.length>0)
        tid_поток_журналирования.send(thisTid, текущее_время(), 
             format("Следующие модули будут журналироваться общим уровнем %s. \n", ошибки_конвертации_уровней_модулей));
    
}

/++
Основная функция записи события в журнал

Params:
    модуль = предполагается имя вызывающего модуля. Влияет только в том случае,
       если такое имя присутствует в массиве параметры_журнала.уровни_модулей в функции инициализации;
       но в любом случае записывается в графу "модуль".
    уровень = уровень события. Если уровень ниже определённого для этого модуля,
       то запись в параметры_журнала не производится.
    запись = любой текст, попадающий в графу "Сообщение"

Examples:
---------------
    в_журнал("модуль1", УровниЖурнала.Предупреждение, "Здесь был Вася");
---------------
+/
void в_журнал(string модуль, УровниЖурнала уровень, lazy string запись) {
    //УровниЖурнала текущий_уровень=параметры_журнала.уровни_модулей.get(модуль, параметры_журнала.текущий_общий_уровень);
    УровниЖурнала текущий_уровень;
    with (параметры_журнала) {
        if ((модуль in уровни_модулей) is null)
            текущий_уровень = текущий_общий_уровень;
        else
            текущий_уровень = уровни_модулей[модуль];
        //УровниЖурнала текущий_уровень=уровни_модулей.get(модуль, текущий_общий_уровень);
        //writeln("текущий_уровень=", текущий_уровень, "  текущий_общий_уровень=", текущий_общий_уровень);
        //writeln("уровни_модулей=", уровни_модулей);
        if (уровень<=текущий_уровень) {
            long микросекунды=текущее_время();
            //float тиков_в_секунду=TickDuration.ticksPerSec;
            float секунды=микросекунды/МНОЖИТЕЛЬ_ВРЕМЕНИ;
            //tid_поток_журналирования.send(thisTid, микросекунды, (format("%10.5f| ", секунды)) ~ модуль ~ "\t|" ~ Soobch_Uroven[уровень] ~ " | " ~ запись ~ "\n");
            locate(ИМЯ_ПОТОКА_ЖУРНАЛИРОВАНИЯ).send(thisTid, микросекунды, 
			  (format("%10.5f|", секунды)) ~ format(формат_имени_модуля, модуль) ~ 
			     "|" ~ УровеньСообщения[уровень] ~ " | " ~ запись ~ "\n");
        }
    }
}

/++
Форматирование многострочных записей для красоты журнала.

Если в журнал отправляется длинный многострочный текст в качестве записи,
то его лучше перед отправкой в журнал пропустить через эту функцию,
тогда все строки будут аккуратно располагаться друг под другом в графе "Сообщение".

Params:
    запись = текст, который надо отправить в журнал.

Returns:
    Текст с отступами в начале строк, пригодный для отправки в журнал.

+/
string журнал_многострочное_форматирование(string запись) {
	auto строки=lineSplitter(запись);
	string возвращаемая_строка=строки.join(параметры_журнала.сдвиг_записей);
	return возвращаемая_строка;
}

/++
Функция, принудительно сбрасывающая информацию в файл журнала и
отключающая дальнейшее журналирование. 

Обычно вызывать не требуется, при выходе из программы (или если 
в ней случилась ошибка) журнал корректно сохраняется. Функция добавлена 
для целей тестирования.
+/
void остановить_журнал() {
    enum long КОД_ВОЗВРАТА = -1;
    auto поток = locate(ИМЯ_ПОТОКА_ЖУРНАЛИРОВАНИЯ);
    if (поток != Tid.init) 
        поток.send(thisTid, КОД_ВОЗВРАТА, "стоп");
}

//-----------------------------------------------------------
// Внутренняя часть 
//------------------------------------------------------------


// --- Формирование заголовка журнала -----
// --- дополнительные действия - инициализация переменных 
// --- параметры_журнала.формат_имени_модуля и параметры_журнала.сдвиг_записей    
private string заголовок() {

	enum ШИРИНА_ИМЕНИ_МОДУЛЯ=15;	

	enum СТРОКА1="Начало работы журнала\n";
	enum разделитель="-".repeat(60+ШИРИНА_ИМЕНИ_МОДУЛЯ).join() ~ "\n";
	enum СТРОКА21="   Время  |  ";
	enum СТРОКА22="Модуль";
	enum СТРОКА23="|  Уровень   | ";
	enum СТРОКА24=" ".repeat(10).join() ~ "Сообщение \n";
	
	// попробовать поставить везде enum вместо string
	enum формат_заголовка_модулей="%-" ~ format("%s", ШИРИНА_ИМЕНИ_МОДУЛЯ-2) ~ "s"; 
	//writeln(формат_имени_модуля);
	enum заголовок_модулей=format (формат_заголовка_модулей, СТРОКА22);
	enum строка2=СТРОКА21 ~ заголовок_модулей ~ СТРОКА23;
	enum формат_пустоты= "%-" ~ format("%s", std.utf.count(строка2)-2) ~ "s";
	параметры_журнала.формат_имени_модуля="%-" ~ format("%s", ШИРИНА_ИМЕНИ_МОДУЛЯ) ~ "s";
	параметры_журнала.сдвиг_записей="\n" ~ format(формат_пустоты, " ") ~ "| ";
	//writeln(параметры_журнала.сдвиг_записей, строка2.length, " ", std.utf.count(строка2));
    //writeln("параметры_журнала=", параметры_журнала);
	return СТРОКА1 ~ разделитель ~ строка2 ~ СТРОКА24 ~ разделитель;
}

private enum УровниЖурнала[string] ПреобразовательУровнейЖурнала = [
    "КритическаяОшибка" : УровниЖурнала.КритическаяОшибка,
    "Ошибка"            : УровниЖурнала.Ошибка,
    "Предупреждение"    : УровниЖурнала.Предупреждение,
    "Информация"        : УровниЖурнала.Информация,
    "Отладка"           : УровниЖурнала.Отладка        
    ];
    
// --- Структура с переменными, расшаренными на все потоки ---	
private struct ПараметрыЖурнала {
    УровниЖурнала текущий_общий_уровень;
    УровниЖурнала[string] уровни_модулей;
    //float тиков_в_секунду;
	string формат_имени_модуля;	// Нужно для правильного расположения названия модуля в функции в_журнал
	string сдвиг_записей; 		// Нужно для функции logformatmultiline
}; 

// --- Экземпляр этой структуры ----
shared private ПараметрыЖурнала параметры_журнала;

private Tid tid_поток_журналирования;


private class ЖурналОбъект {
    File файл_журнала;
    long[]  ключи;
    string[long] строки;    
    int размер_буфера=10, 
        размер_выгрузки=5,
        занято=0;
    
    this(string имя_файла_журнала, int размер_буфера, int размер_выгрузки) {
        this.размер_буфера=размер_буфера;
        if (размер_выгрузки>размер_буфера)
            this.размер_выгрузки=размер_буфера;
        else 
            this.размер_выгрузки=размер_выгрузки;
        ключи= new long[размер_буфера+1];
        //writeln("присвоили переменные в структуре");
        файл_журнала=File(имя_файла_журнала, "w");
        //writeln("открыли файл");
        строки=[0L:""];
        //writeln("присвоили строки");
        //строки=строки1;
    }
    
    ~this() {
        записать (занято);
        файл_журнала.close();
    }
    
    void в_журнал(long время, string строка) {
        if (занято==0) 
            ключи[занято] = время;
        else
            if (ключи[занято-1] < время)
                ключи[занято] = время;
            else
                вставить_ключ(время);
        занято++;
        строки[время]=строка;
        if (занято>=размер_буфера) записать(размер_выгрузки);
    }
    
    void вставить_ключ(long время) {
        //writeln("занято=", занято);
        //writeln("время=", время);
        //writeln("ключи=", ключи);
        if (занято==размер_буфера)
            writeln("переполнение буфера. ");
        int i;
        
        for (i=занято-1; i>=0; i--) //{
            if (ключи[i]>время)
                ключи[i+1]=ключи[i];
            else 
                break;
        ключи[i+1]=время;
        //writeln("ключи=", ключи);
    }
    
    void записать (int размер) {
        if (размер>занято) размер=занято;
        string буфер="";
        foreach (время; ключи[0..размер]) {
            буфер~=строки[время];
            строки.remove(время);
        }
        //буфер~="--- Записан буфер ----\n";
        файл_журнала.write(буфер);
        for (int i=размер, j=0; i<занято; i++, j++)
            ключи[j]=ключи[i];
        занято=занято-размер;
    }
}

private void функция_потока_журнала(string имя_файла_журнала, int размер_буфера, int размер_выгрузки) {
    ЖурналОбъект объект_журнала=new ЖурналОбъект(имя_файла_журнала, размер_буфера, размер_выгрузки);
    try {
        //ЖурналОбъект объект_журнала=ЖурналОбъект(имя_файла_журнала, размер_буфера, размер_выгрузки);
        for (;;) {
            auto msg = receiveOnly!(Tid, long, string)();
            if (msg[1] < 0) {
                //writeln("поток остановили");
                break;
            }
            объект_журнала.в_журнал(msg[1], msg[2]);
        }
    } catch (OwnerTerminated exc) {
        //destroy(объект_журнала); 
    } catch (MessageMismatch exc) {
        writeln("случилась ошибка MessageMismatch");
        //destroy(объект_журнала); 
    } 
    
    destroy(объект_журнала); 
    
 }


shared static this() {
    время_старта = MonoTime.currTime;
}


immutable private MonoTime время_старта;
// Возвращает время после запуска программы в микросекундах
private long текущее_время() {
	return (MonoTime.currTime - время_старта).total!"hnsecs";
}

 
