module my_utils.nastroyki_struct;

import std.stdio, std.file, std.string, std.conv;
import std.range: iota;
public import std.meta: AliasSeq;
import std.algorithm.iteration: splitter;


/++
Структура, определяющая настройки пользовательской программы.

Для создания экземпляра структуры необходимо определить кортеж $(B std.meta.AliasSeq) следующего формата: $(BR)
---
alias Кортеж_Настроек = AliasSeq!( 
  //Тип       наименование свойства   значение по-умолчанию
    long,           "параметр1",        "0",
    string,         "параметр2",      "\"строка\"",
    ...
    )
---
Тип кортежа AliasSeq необязательно импортировать отдельно, он импортирован публично в этом модуле.

Допустимые типы для свойств структуры $(B Настройки):
---
long, double, string, long[], double[], string[], string[string]
---
Строковые значения необходимо обрамлять экранированными кавычками \" (как в примере вверху).

Загрузка настроек из файла осуществляется методом 
$(LINK2 #Настройки.прочитать_настройки_из_файла, прочитать_настройки_из_файла).
Отсутствующие в файле свойства (или те, что загружались с ошибкой) остаются со значением по-умолчанию.
Доступ к свойству осуществляется через @property-метод с именем свойства:
---
long x = настройки.параметр1;
---
Свойства реализованы в режиме только для чтения.

+/
struct Настройки(Кортеж_Настроек...) {
    /*
    private double[string] настройки_double;
    private long[string] настройки_long;
    private string[string] настройки_string;
    private Массив_string[string] настройки_массив_string;
    private Массив_long[string] настройки_массив_long;
    private Массив_double[string] настройки_массив_double;
    private Ассоциативный_Массив_string[string] настройки_ассоциативный_массив;
    */
    
    // Cюда записываются имена свойств, присутствующих в файле с настройками, но при
    // получении значений которых случились ошибки преобразования типа

    private string[] _ошибочные_переменные; 

    /++
    Возвращает имена свойств, присутствующих в файле с настройками, но при
    получении значений которых случились ошибки преобразования типа.
    +/
    string[] ошибочные_переменные() @property { return _ошибочные_переменные; } 

    // Становится равным true, если успешно загружено значение свойства
    private bool _загружено = false; //

    /++
    Произошла ли успешная загрузка хотя бы одного значения свойства.
    +/
    bool загружено() @property { return _загружено; } 


    /++
    Загрузка настроек из файла.

    В случае отсутствия файла или ошибки чтения выводится соответствующее 
    сообщение, а свойства остаются со значениями по-умолчанию.
    
    Params:
        имя_файла = файл, из которого нужно прочитать значения настроек.
        разделитель = символ, которым разделяются строки со свойствами; допустимы символы '\n' и ';'.
    +/
    bool прочитать_настройки_из_файла(string имя_файла, char разделитель='\n') { 
        string текст = "";
        if (exists(имя_файла)) {
            try 
                текст = readText(имя_файла);
            catch (Exception) 
                writeln("Файл с настройками ", имя_файла, " не читается!");
        }
        else 
             writeln("Не найден файл с настройками ", имя_файла);
        return прочитать_настройки(текст, разделитель); 
    }


    /++
    Загрузка настроек из строки с текстом. 

    Метод определён для целей тестирования, обычно в структуре $(B Настройки)
    используется метод $(LINK2 #Настройки.прочитать_настройки_из_файла, прочитать_настройки_из_файла).

    Строки со значениями свойств определяются в виде $(BR)
    $(B имя_свойства = значение) $(BR)
    Пустые строки, и строки, начинающиеся с символа '#', игнорируются.
    Также игнорируются строки, имя свойства в которых отсутствует в параметре шаблона Кортеж_Настроек
        
    Params:
        текст = текст, из которого нужно прочитать настройки.
        разделитель = символ, которым разделяются строки со свойствами; допустимы символы '\n' и ';'.

    Note:
        В случае, если при преобразовании типа существующего свойства произошла ошибка,
        имя этого свойства записывается в массив $(B ошибочные_переменные), а у самого
        свойства остаётся значение по-умолчанию.
    +/
    bool прочитать_настройки(string текст, char разделитель='\n') { //, string[][string] имена_всех_настроек) {
        if (this.toString == "")    // При создании инициализировать не получается (ошибка компиляции)
            this.инициализация();   // Проводим инициализацию перед прочтением файла


        auto текст1 = splitter(текст, разделитель);
        string[string] текст2;
        string имя_настройки, значение_настройки;
        foreach(string строка1; текст1) {
            //writeln(строка1);
            string строка2 = stripLeft(строка1);
            if (строка2.length<2)
                continue;
            if (строка2[0..1]=="#")
                continue;
            auto индекс_равно = indexOf(строка1, "=");
            if (индекс_равно<0) 
                continue;
            имя_настройки = очистка(строка1[0 .. индекс_равно]);
            значение_настройки = очистка(строка1[индекс_равно+1 .. $]);
            текст2[имя_настройки] = значение_настройки;
        }
        //writeln("текст2=",текст2);
        _ошибочные_переменные = [];

        присвоение_значений(текст2);
        //чтение(текст2, настройки_double); 

        /*
        чтение!double(текст2, настройки_double); 
        чтение!long(текст2, настройки_long); 
        чтение!string(текст2, настройки_string); 
        чтение!Массив_string(текст2, настройки_массив_string); 
        чтение!Массив_long(текст2, настройки_массив_long); 
        чтение!Массив_double(текст2, настройки_массив_double); 
        чтение!Ассоциативный_Массив_string(текст2, настройки_ассоциативный_массив); 
        */
        return _загружено;
    }


    private void чтение(T)(string[string] текст2, ref T[string] настройки_T) {
        foreach(string имя_настройки; имена_свойств) { 
            //string имя_настройки = Кортеж_Настроек[i*3+1];
            if ((имя_настройки in текст2) && (имя_настройки in настройки_T)) {
                try {
                    T значение = to!(T)(текст2[имя_настройки]);
                    настройки_T[имя_настройки] = значение;
                    _загружено = true;
                }
                catch (ConvException) {
                    _ошибочные_переменные ~= имя_настройки;
                    //writeln("ConvException. имя_настройки=", имя_настройки, "  значение=", текст2[имя_настройки]);
                    }
            }
            //else
            //    _ошибочные_переменные ~= имя_настройки;
        }
    }       

    /++
    Построчный вывод значений всех свойств структуры.
    +/    
    public string toString() {
        return _toString();
        /*
        string результат="";
        результат ~= to_string(настройки_double);
        результат ~= to_string(настройки_long);
        результат ~= to_string(настройки_string);
        результат ~= to_string(настройки_массив_double);
        результат ~= to_string(настройки_массив_long);
        результат ~= to_string(настройки_массив_string);
        результат ~= to_string(настройки_ассоциативный_массив);
        return результат;
        */
    }

    /++
    Возвращает значения строк генерации (для целей отладки). 
    Работает только в debug-режиме, иначе возвращает пустую строку.
    +/    
    public string вывод_строк_генерации() {
        return _вывод_строк_генерации();
    }

    private string to_string(T)(T[string] настройки_T) {
        string результат="";
        foreach(имя_настройки, значение; настройки_T)
            результат ~= имя_настройки ~ format(" = %s \n", значение);
        return результат;
    }
    
    private string очистка(string аргумент) {
        string результат=strip(аргумент);
        if ((результат[0]=='"') && (результат[$-1]=='"'))
            результат = результат[1 .. $-1];
        return результат;
    }

    //В режиме debug генерируются enum-константы сгенерированных методов и членов
  debug {
    private enum строка_генерации_массивов = Генератор_Массивов();
    mixin(строка_генерации_массивов);
    
    private enum строка_генерации_имён = Генератор_Имён!Кортеж_Настроек();
    mixin(строка_генерации_имён);
    
    private enum строка_генерации_свойств = Генератор_Свойств!Кортеж_Настроек();
    mixin(строка_генерации_свойств);

    private enum строка_генерации_инициализатора = Генерация_инициализатора!Кортеж_Настроек();
    mixin(строка_генерации_инициализатора);

    private enum строка_генерации__toString = Генератор__toString();
    mixin(строка_генерации__toString);

    private enum строка_генерации_присвоения_значений = Генератор_присвоения_значений();
    mixin(строка_генерации_присвоения_значений);

    private string _вывод_строк_генерации() {
        string строка = "";
        строка ~= "строка_генерации_массивов =\n" ~ строка_генерации_массивов ~ "\n";
        строка ~= "строка_генерации_свойств =\n" ~ строка_генерации_свойств ~ "\n";
        строка ~= "строка_генерации_имён =\n" ~ строка_генерации_имён ~ "\n\n";
        строка ~= "строка_генерации_инициализатора =\n" ~ строка_генерации_инициализатора ~ "\n";
        строка ~= "строка_генерации__toString =\n" ~ строка_генерации__toString ~ "\n";
        строка ~= "строка_генерации_присвоения_значений =\n" ~ строка_генерации_присвоения_значений;
        return строка;
    }
  }
  else {
    mixin(Генератор_Массивов());
    mixin(Генератор_Имён!Кортеж_Настроек());
    mixin(Генератор_Свойств!Кортеж_Настроек());
    mixin(Генерация_инициализатора!Кортеж_Настроек());
    mixin(Генератор__toString());
    mixin(Генератор_присвоения_значений());
    private string _вывод_строк_генерации() {
      return "";
    }

  }
    
}


//-----------------------------------------------------------
// Внутренняя часть 
//------------------------------------------------------------

// Объявляем типы массивов
private alias Массив_string = string[];
private alias Массив_long = long[];
private alias Массив_double = double[];
private alias Ассоциативный_Массив_string = string[string];

// Кортеж с типами и именами ассоциативных массивов в структуре
// (чтобы можно было перебирать их в циклах функций генераторов миксинов)
private alias Массивы = AliasSeq!(
    //int, "настройки_int",
    long, "настройки_long", 
    double, "настройки_double", 
    string, "настройки_string",
    Массив_string, "настройки_массив_string", 
    Массив_long, "настройки_массив_long", 
    Массив_double, "настройки_массив_double",
    Ассоциативный_Массив_string, "настройки_ассоциативный_массив"
);

// Функция, генерурующая миксин-строку с определениями ассоциативных массивов, заданных в 
// кортеже Массивы
private string Генератор_Массивов() {
    string строка;
    static foreach(j; iota(Массивы.length/2)) 
        строка ~= "private " ~ Массивы[j*2].stringof ~ "[string] " ~ Массивы[j*2+1] ~ "; \n";
    return строка;
}

// Функция, используемая для определения одного свойства в функции-генераторе Генератор_Свойств
private string элемент(string тип, string имя_элемента, string настройки_типа) {
    return тип ~ " " ~ имя_элемента ~ "() @property  { return " ~ настройки_типа ~ "[\"" ~ имя_элемента ~ "\"]; }\n";
}

// Функция, генерурующая миксин-строку с определениями свойств, заданных в 
// кортеже Кортеж, передаваемом параметром шаблона
private string Генератор_Свойств(Кортеж...)() {
    string строка;
    static foreach(i; iota(Кортеж.length/3)) {
        static foreach(j; iota(Массивы.length/2)) {
            static if (is(Кортеж[i*3] == Массивы[j*2]))
                строка ~= элемент(Кортеж[i*3].stringof, Кортеж[i*3+1], Массивы[j*2+1]);
        }
    }
    return строка;
}

// Функция, генерирующая массив имён свойств
// (массив используется в методе структуры чтение!T())
private string Генератор_Имён(Кортеж...)() {
    string строка = "private enum имена_свойств = [";
    static foreach(i; iota(Кортеж.length/3)) 
        строка ~= "\"" ~ Кортеж[i*3+1] ~ "\",";
    строка ~= "];";
    return строка;
}

// Функция, используемая для инициализации одного значения
// в функции-генераторе Генерация_инициализатора
private string присвоение_значения(string имя_массива, string имя_элемента, string значение) {
    return имя_массива ~ "[\"" ~ имя_элемента ~ "\"] = " ~ значение ~ ";\n";
} 

// Функция, генерирующая метод инициализация(), инициализирующий значениями
// по-умолчанию всех пар ключ-значение ассоциативных массивов настройки_тип,
// в соответствии с содержимым кортежа Кортеж, передаваемом параметром шаблона
private string Генерация_инициализатора(Кортеж...)() {
    string строка = "void инициализация() {\n";
    static foreach(i; iota(Кортеж.length/3)) 
        static foreach(j; iota(Массивы.length/2)) {
            static if (is(Кортеж[i*3] == Массивы[j*2]))
                строка ~= присвоение_значения(Массивы[j*2+1], Кортеж[i*3+1], Кортеж[i*3+2]);
        }
    строка ~= "}\n";
    return строка;
}

// Функция, генерирующая метод _toString(), выводящий все свойства в
// построчно оформленный текст
private string Генератор__toString() {
    string строка = "private string _toString() {\nstring результат=\"\";\n";
    static foreach(j; iota(Массивы.length/2)) 
        строка ~= "результат ~= to_string(" ~ Массивы[j*2+1] ~ "); \n";
    строка ~= "return результат;\n}\n";
    return строка;
}

// Функция, генерирующая метод присвоение_значений(), присваивающий загруженные 
// из файла значения в элементам массивов
private string Генератор_присвоения_значений() {
    string строка = "private void присвоение_значений(string[string] исходные_данные) {\n";
    static foreach(j; iota(Массивы.length/2)) 
        строка ~= "чтение!("  ~ Массивы[j*2].stringof ~ ")(исходные_данные, " ~ Массивы[j*2+1] ~ "); \n";
    строка ~= "}\n";
    return строка;
}

